"""
Photon emission and absorption models for use with the
photon simulator.
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.utilities.on_demand_imports import _h5py as h5py
import numpy as np
import os

from yt.funcs import mylog
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.on_demand_imports import _astropy
from yt.utilities.physical_constants import hcgs, clight
from yt.utilities.physical_ratios import erg_per_keV, amu_grams
from yt.analysis_modules.photon_simulator.utils import broaden_lines

hc = (hcgs*clight).in_units("keV*angstrom").v
cl = clight.v
K = 1.0/np.sqrt(2.*np.pi)

class SpectralModel(object):

    def __init__(self, emin, emax, nchan):
        self.emin = YTQuantity(emin, "keV")
        self.emax = YTQuantity(emax, "keV")
        self.nchan = nchan
        self.ebins = YTArray(np.linspace(self.emin, self.emax, nchan+1), "keV")
        self.de = np.diff(self.ebins)
        self.emid = 0.5*(self.ebins[1:]+self.ebins[:-1])

    def prepare_spectrum(self):
        pass

    def get_spectrum(self):
        pass

    def cleanup_spectrum(self):
        pass

class XSpecThermalModel(SpectralModel):
    r"""
    Initialize a thermal gas emission model from PyXspec.

    Parameters
    ----------
    model_name : string
        The name of the thermal emission model.
    emin : float
        The minimum energy for the spectral model.
    emax : float
        The maximum energy for the spectral model.
    nchan : integer
        The number of channels in the spectral model.
    settings : dictionary, optional
        A dictionary of key, value pairs (must both be strings)
        that can be used to set various options in XSPEC.

    Examples
    --------
    >>> mekal_model = XSpecThermalModel("mekal", 0.05, 50.0, 1000)
    """
    def __init__(self, model_name, emin, emax, nchan,
                 thermal_broad=False, settings=None):
        self.model_name = model_name
        self.thermal_broad = thermal_broad
        if settings is None: settings = {}
        self.settings = settings
        super(XSpecThermalModel, self).__init__(emin, emax, nchan)

    def prepare_spectrum(self, zobs):
        """
        Prepare the thermal model for execution given a redshift *zobs* for the spectrum.
        """
        import xspec
        xspec.Xset.chatter = 0
        xspec.AllModels.setEnergies("%f %f %d lin" %
                                    (self.emin.value, self.emax.value, self.nchan))
        self.model = xspec.Model(self.model_name)
        self.thermal_comp = getattr(self.model,self.model_name)
        if self.model_name == "bremss":
            self.norm = 3.02e-15
        else:
            self.norm = 1.0e-14
        self.thermal_comp.norm = 1.0
        self.thermal_comp.Redshift = zobs
        if self.thermal_broad:
            xspec.Xset.addModelString("APECTHERMAL","yes")
        for k,v in self.settings.items():
            xspec.Xset.addModelString(k,v)

    def get_spectrum(self, kT):
        """
        Get the thermal emission spectrum given a temperature *kT* in keV. 
        """
        self.thermal_comp.kT = kT
        self.thermal_comp.Abundanc = 0.0
        cosmic_spec = np.array(self.model.values(0))
        if self.model_name == "bremss":
            metal_spec = np.zeros(self.nchan)
        else:
            self.thermal_comp.Abundanc = 1.0
            metal_spec = np.array(self.model.values(0)) - cosmic_spec
        cosmic_spec *= self.norm
        metal_spec *= self.norm
        return YTArray(cosmic_spec, "cm**3/s"), YTArray(metal_spec, "cm**3/s")
    
    def cleanup_spectrum(self):
        del self.thermal_comp
        del self.model

class XSpecAbsorbModel(SpectralModel):
    r"""
    Initialize an absorption model from PyXspec.

    Parameters
    ----------
    model_name : string
        The name of the absorption model.
    nH : float
        The foreground column density *nH* in units of 10^22 cm^{-2}.
    emin : float, optional
        The minimum energy for the spectral model.
    emax : float, optional
        The maximum energy for the spectral model.
    nchan : integer, optional
        The number of channels in the spectral model.
    settings : dictionary, optional
        A dictionary of key, value pairs (must both be strings)
        that can be used to set various options in XSPEC.

    Examples
    --------
    >>> abs_model = XSpecAbsorbModel("wabs", 0.1)
    """
    def __init__(self, model_name, nH, emin=0.01, emax=50.0,
                 nchan=100000, settings=None):
        self.model_name = model_name
        self.nH = nH
        if settings is None: settings = {}
        self.settings = settings
        super(XSpecAbsorbModel, self).__init__(emin, emax, nchan)

    def prepare_spectrum(self):
        """
        Prepare the absorption model for execution given a redshift *zobs* for the spectrum.
        """
        import xspec
        xspec.Xset.chatter = 0
        xspec.AllModels.setEnergies("%f %f %d lin" %
                                    (self.emin.value, self.emax.value, self.nchan))
        self.model = xspec.Model(self.model_name+"*powerlaw")
        self.model.powerlaw.norm = self.nchan/(self.emax.value-self.emin.value)
        self.model.powerlaw.PhoIndex = 0.0
        for k,v in self.settings.items():
            xspec.Xset.addModelString(k,v)

    def get_spectrum(self):
        """
        Get the absorption spectrum.
        """
        m = getattr(self.model,self.model_name)
        m.nH = self.nH
        return np.array(self.model.values(0))

    def cleanup_spectrum(self):
        del self.model

class TableApecModel(SpectralModel):
    r"""
    Initialize a thermal gas emission model from the AtomDB APEC tables
    available at http://www.atomdb.org. This code borrows heavily from Python
    routines used to read the APEC tables developed by Adam Foster at the
    CfA (afoster@cfa.harvard.edu).

    Parameters
    ----------
    apec_root : string
        The directory root where the APEC model files are stored.
    emin : float
        The minimum energy for the spectral model.
    emax : float
        The maximum energy for the spectral model.
    nchan : integer
        The number of channels in the spectral model.
    apec_vers : string, optional
        The version identifier string for the APEC files, e.g.
        "2.0.2"
    thermal_broad : boolean, optional
        Whether to apply thermal broadening to spectral lines. Only should
        be used if you are attempting to simulate a high-spectral resolution
        detector.

    Examples
    --------
    >>> apec_model = TableApecModel("$SPECTRAL_DATA/spectral/", 0.05, 50.0,
    ...                             1000, apec_vers="3.0", thermal_broad=True)
    """
    def __init__(self, apec_root, emin, emax, nchan,
                 apec_vers="2.0.2", thermal_broad=False):
        self.apec_root = apec_root
        self.apec_prefix = "apec_v"+apec_vers
        self.cocofile = os.path.join(self.apec_root,
                                     self.apec_prefix+"_coco.fits")
        self.linefile = os.path.join(self.apec_root,
                                     self.apec_prefix+"_line.fits")
        super(TableApecModel, self).__init__(emin, emax, nchan)
        self.wvbins = hc/self.ebins[::-1].d
        # H, He, and trace elements
        self.cosmic_elem = [1,2,3,4,5,9,11,15,17,19,21,22,23,24,25,27,29,30]
        # Non-trace metals
        self.metal_elem = [6,7,8,10,12,13,14,16,18,20,26,28]
        self.thermal_broad = thermal_broad
        self.A = np.array([0.0,1.00794,4.00262,6.941,9.012182,10.811,
                           12.0107,14.0067,15.9994,18.9984,20.1797,
                           22.9898,24.3050,26.9815,28.0855,30.9738,
                           32.0650,35.4530,39.9480,39.0983,40.0780,
                           44.9559,47.8670,50.9415,51.9961,54.9380,
                           55.8450,58.9332,58.6934,63.5460,65.3800])

    def prepare_spectrum(self, zobs):
        """
        Prepare the thermal model for execution.
        """
        try:
            self.line_handle = _astropy.pyfits.open(self.linefile)
        except IOError:
            mylog.error("LINE file %s does not exist" % self.linefile)
            raise IOError("LINE file %s does not exist" % self.linefile)
        try:
            self.coco_handle = _astropy.pyfits.open(self.cocofile)
        except IOError:
            mylog.error("COCO file %s does not exist" % self.cocofile)
            raise IOError("COCO file %s does not exist" % self.cocofile)

        self.Tvals = self.line_handle[1].data.field("kT")
        self.dTvals = np.diff(self.Tvals)
        self.minlam = self.wvbins.min()
        self.maxlam = self.wvbins.max()
        self.scale_factor = 1.0/(1.+zobs)

    def _make_spectrum(self, kT, element, tindex):

        tmpspec = np.zeros(self.nchan)

        line_data = self.line_handle[tindex].data
        coco_data = self.coco_handle[tindex].data

        i = np.where((line_data.field('element') == element) &
                     (line_data.field('lambda') > self.minlam) &
                     (line_data.field('lambda') < self.maxlam))[0]

        E0 = hc/line_data.field('lambda')[i].astype("float64")*self.scale_factor
        amp = line_data.field('epsilon')[i].astype("float64")
        ebins = self.ebins.d
        de = self.de.d
        emid = self.emid.d
        if self.thermal_broad:
            sigma = E0*np.sqrt(2.*kT*erg_per_keV/(self.A[element]*amu_grams))/cl
            vec = broaden_lines(E0, sigma, amp, ebins)
        else:
            vec = np.histogram(E0, ebins, weights=amp)[0]
        tmpspec += vec

        ind = np.where((coco_data.field('Z') == element) &
                       (coco_data.field('rmJ') == 0))[0]
        if len(ind) == 0:
            return tmpspec
        else:
            ind = ind[0]

        n_cont = coco_data.field('N_Cont')[ind]
        e_cont = coco_data.field('E_Cont')[ind][:n_cont]
        continuum = coco_data.field('Continuum')[ind][:n_cont]

        tmpspec += np.interp(emid, e_cont*self.scale_factor, continuum)*de/self.scale_factor

        n_pseudo = coco_data.field('N_Pseudo')[ind]
        e_pseudo = coco_data.field('E_Pseudo')[ind][:n_pseudo]
        pseudo = coco_data.field('Pseudo')[ind][:n_pseudo]

        tmpspec += np.interp(emid, e_pseudo*self.scale_factor, pseudo)*de/self.scale_factor

        return tmpspec*self.scale_factor

    def get_spectrum(self, kT):
        """
        Get the thermal emission spectrum given a temperature *kT* in keV. 
        """
        cspec_l = np.zeros(self.nchan)
        mspec_l = np.zeros(self.nchan)
        cspec_r = np.zeros(self.nchan)
        mspec_r = np.zeros(self.nchan)
        tindex = np.searchsorted(self.Tvals, kT)-1
        if tindex >= self.Tvals.shape[0]-1 or tindex < 0:
            return YTArray(cspec_l, "cm**3/s"), YTArray(mspec_l, "cm**3/s")
        dT = (kT-self.Tvals[tindex])/self.dTvals[tindex]
        # First do H,He, and trace elements
        for elem in self.cosmic_elem:
            cspec_l += self._make_spectrum(kT, elem, tindex+2)
            cspec_r += self._make_spectrum(kT, elem, tindex+3)
        # Next do the metals
        for elem in self.metal_elem:
            mspec_l += self._make_spectrum(kT, elem, tindex+2)
            mspec_r += self._make_spectrum(kT, elem, tindex+3)
        cosmic_spec = YTArray(cspec_l*(1.-dT)+cspec_r*dT, "cm**3/s")
        metal_spec = YTArray(mspec_l*(1.-dT)+mspec_r*dT, "cm**3/s")
        return cosmic_spec, metal_spec

class TableAbsorbModel(SpectralModel):
    r"""
    Initialize an absorption model from a table stored in an HDF5 file.

    Parameters
    ----------
    filename : string
        The name of the table file.
    nH : float
        The foreground column density *nH* in units of 10^22 cm^{-2}.

    Examples
    --------
    >>> abs_model = XSpecAbsorbModel("abs_table.h5", 0.1)
    """
    def __init__(self, filename, nH):
        if not os.path.exists(filename):
            raise IOError("File does not exist: %s." % filename)
        self.filename = filename
        f = h5py.File(self.filename,"r")
        emin = f["energy"][:].min()
        emax = f["energy"][:].max()
        self.sigma = YTArray(f["cross_section"][:], "cm**2")
        nchan = self.sigma.shape[0]
        f.close()
        super(TableAbsorbModel, self).__init__(emin, emax, nchan)
        self.nH = YTQuantity(nH*1.0e22, "cm**-2")

    def prepare_spectrum(self):
        """
        Prepare the absorption model for execution.
        """
        pass

    def get_spectrum(self):
        """
        Get the absorption spectrum.
        """
        return np.exp(-self.sigma*self.nH)



"""
Particle trajectories
"""
from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.data_objects.field_data import YTFieldData
from yt.utilities.lib.particle_mesh_operations import CICSample_3
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_root_only
from yt.funcs import mylog, get_pbar
from yt.units.yt_array import array_like_field
from yt.config import ytcfg
from yt.utilities.exceptions import YTIllDefinedParticleData
from collections import OrderedDict

import numpy as np
from yt.utilities.on_demand_imports import _h5py as h5py

class ParticleTrajectories(object):
    r"""A collection of particle trajectories in time over a series of
    datasets.

    Parameters
    ----------
    outputs : ~yt.data_objects.time_series.DatasetSeries
        DatasetSeries object from which to draw the particles.
    indices : array_like
        An integer array of particle indices whose trajectories we
        want to track. If they are not sorted they will be sorted.
    fields : list of strings, optional
        A set of fields that is retrieved when the trajectory
        collection is instantiated. Default: None (will default
        to the fields 'particle_position_x', 'particle_position_y',
        'particle_position_z')
    suppress_logging : boolean
        Suppress yt's logging when iterating over the simulation time
        series. Default: False
    ptype : str, optional
        Only use this particle type. Default: None, which uses all particle type.

    Examples
    --------
    >>> my_fns = glob.glob("orbit_hdf5_chk_00[0-9][0-9]")
    >>> my_fns.sort()
    >>> fields = ["particle_position_x", "particle_position_y",
    >>>           "particle_position_z", "particle_velocity_x",
    >>>           "particle_velocity_y", "particle_velocity_z"]
    >>> ds = load(my_fns[0])
    >>> init_sphere = ds.sphere(ds.domain_center, (.5, "unitary"))
    >>> indices = init_sphere["particle_index"].astype("int")
    >>> ts = DatasetSeries(my_fns)
    >>> trajs = ts.particle_trajectories(indices, fields=fields)
    >>> for t in trajs :
    >>>     print t["particle_velocity_x"].max(), t["particle_velocity_x"].min()
    """
    def __init__(self, outputs, indices, fields=None, suppress_logging=False, ptype=None):

        indices.sort() # Just in case the caller wasn't careful
        self.field_data = YTFieldData()
        self.data_series = outputs
        self.masks = []
        self.sorts = []
        self.array_indices = []
        self.indices = indices
        self.num_indices = len(indices)
        self.num_steps = len(outputs)
        self.times = []
        self.suppress_logging = suppress_logging
        self.ptype = ptype

        if fields is None: fields = []
        fields = list(OrderedDict.fromkeys(fields))

        if self.suppress_logging:
            old_level = int(ytcfg.get("yt","loglevel"))
            mylog.setLevel(40)
        ds_first = self.data_series[0]
        dd_first = ds_first.all_data()

        fds = {}
        for field in (
                "particle_index",
                "particle_position_x", "particle_position_y", "particle_position_z"):
            fds[field] = self._get_full_field_name(field)[0]

        my_storage = {}
        pbar = get_pbar("Constructing trajectory information", len(self.data_series))
        for i, (sto, ds) in enumerate(self.data_series.piter(storage=my_storage)):
            dd = ds.all_data()
            newtags = dd[fds["particle_index"]].d.astype("int64")
            mask = np.in1d(newtags, indices, assume_unique=True)
            sort = np.argsort(newtags[mask])
            array_indices = np.where(np.in1d(indices, newtags, assume_unique=True))[0]
            self.array_indices.append(array_indices)
            self.masks.append(mask)
            self.sorts.append(sort)

            pfields = {}
            for field in ("particle_position_%s" % ax for ax in "xyz"):
                pfields[field] = dd[fds[field]].ndarray_view()[mask][sort]

            sto.result_id = ds.parameter_filename
            sto.result = (ds.current_time, array_indices, pfields)
            pbar.update(i)
        pbar.finish()

        if self.suppress_logging:
            mylog.setLevel(old_level)

        times = []
        for fn, (time, indices, pfields) in sorted(my_storage.items()):
            times.append(time)
        self.times = self.data_series[0].arr([time for time in times], times[0].units)

        self.particle_fields = []
        output_field = np.empty((self.num_indices, self.num_steps))
        output_field.fill(np.nan)
        for field in ("particle_position_%s" % ax for ax in "xyz"):
            for i, (fn, (time, indices, pfields)) in enumerate(sorted(my_storage.items())):
                try:
                    # This will fail if particles ids are
                    # duplicate. This is due to the fact that the rhs
                    # would then have a different shape as the lhs
                    output_field[indices, i] = pfields[field]
                except ValueError:
                    raise YTIllDefinedParticleData("This dataset contains duplicate particle indices!")
            self.field_data[field] = array_like_field(
                dd_first, output_field.copy(), fds[field])
            self.particle_fields.append(field)

        # Instantiate fields the caller requested
        self._get_data(fields)

    def has_key(self, key):
        return (key in self.field_data)

    def keys(self):
        return self.field_data.keys()

    def _get_full_field_name(self, field):
        ds_first = self.data_series[0]
        dd_first = ds_first.all_data()
        ptype = self.ptype if self.ptype else 'all'
        return dd_first._determine_fields((ptype, field))

    def __getitem__(self, key):
        """
        Get the field associated with key.
        """
        if key == "particle_time":
            return self.times
        if key not in self.field_data:
            self._get_data([key])
        return self.field_data[key]

    def __setitem__(self, key, val):
        """
        Sets a field to be some other value.
        """
        self.field_data[key] = val

    def __delitem__(self, key):
        """
        Delete the field from the trajectory
        """
        del self.field_data[key]

    def __iter__(self):
        """
        This iterates over the trajectories for
        the different particles, returning dicts
        of fields for each trajectory
        """
        for idx in range(self.num_indices):
            traj = {}
            traj["particle_index"] = self.indices[idx]
            traj["particle_time"] = self.times
            for field in self.field_data.keys():
                traj[field] = self[field][idx,:]
            yield traj

    def __len__(self):
        """
        The number of individual trajectories
        """
        return self.num_indices

    def add_fields(self, fields):
        """
        Add a list of fields to an existing trajectory

        Parameters
        ----------
        fields : list of strings
            A list of fields to be added to the current trajectory
            collection.

        Examples
        ________
        >>> trajs = ParticleTrajectories(my_fns, indices)
        >>> trajs.add_fields(["particle_mass", "particle_gpot"])
        """
        self._get_data(fields)

    def _get_data(self, fields):
        """
        Get a list of fields to include in the trajectory collection.
        The trajectory collection itself is a dict of 2D numpy arrays,
        with shape (num_indices, num_steps)
        """

        missing_fields = [field for field in fields
                          if field not in self.field_data]
        if not missing_fields:
            return

        if self.suppress_logging:
            old_level = int(ytcfg.get("yt","loglevel"))
            mylog.setLevel(40)
        ds_first = self.data_series[0]
        dd_first = ds_first.all_data()

        fds = {}
        new_particle_fields = []
        for field in missing_fields:
            fds[field] = dd_first._determine_fields(field)[0]
            if field not in self.particle_fields:
                if self.data_series[0]._get_field_info(*fds[field]).particle_type:
                    self.particle_fields.append(field)
                    new_particle_fields.append(field)


        grid_fields = [field for field in missing_fields
                       if field not in self.particle_fields]
        step = int(0)
        pbar = get_pbar("Generating [%s] fields in trajectories" %
                        ", ".join(missing_fields), self.num_steps)
        my_storage = {}

        for i, (sto, ds) in enumerate(self.data_series.piter(storage=my_storage)):
            mask = self.masks[i]
            sort = self.sorts[i]
            pfield = {}

            if new_particle_fields:  # there's at least one particle field
                dd = ds.all_data()
                for field in new_particle_fields:
                    # This is easy... just get the particle fields
                    pfield[field] = dd[fds[field]].d[mask][sort]

            if grid_fields:
                # This is hard... must loop over grids
                for field in grid_fields:
                    pfield[field] = np.zeros(self.num_indices)
                x = self["particle_position_x"][:,step].d
                y = self["particle_position_y"][:,step].d
                z = self["particle_position_z"][:,step].d
                particle_grids, particle_grid_inds = ds.index._find_points(x,y,z)

                # This will fail for non-grid index objects
                for grid in particle_grids:
                    cube = grid.retrieve_ghost_zones(1, grid_fields)
                    for field in grid_fields:
                        CICSample_3(x, y, z, pfield[field],
                                    self.num_indices,
                                    cube[fds[field]],
                                    np.array(grid.LeftEdge).astype(np.float64),
                                    np.array(grid.ActiveDimensions).astype(np.int32),
                                    grid.dds[0])
            sto.result_id = ds.parameter_filename
            sto.result = (self.array_indices[i], pfield)
            pbar.update(step)
            step += 1
        pbar.finish()

        output_field = np.empty((self.num_indices,self.num_steps))
        output_field.fill(np.nan)
        for field in missing_fields:
            fd = fds[field]
            for i, (fn, (indices, pfield)) in enumerate(sorted(my_storage.items())):
                output_field[indices, i] = pfield[field]
            self.field_data[field] = array_like_field(dd_first, output_field.copy(), fd)

        if self.suppress_logging:
            mylog.setLevel(old_level)

    def trajectory_from_index(self, index):
        """
        Retrieve a single trajectory corresponding to a specific particle
        index

        Parameters
        ----------
        index : int
            This defines which particle trajectory from the
            ParticleTrajectories object will be returned.

        Returns
        -------
        A dictionary corresponding to the particle's trajectory and the
        fields along that trajectory

        Examples
        --------
        >>> from yt.mods import *
        >>> import matplotlib.pylab as pl
        >>> trajs = ParticleTrajectories(my_fns, indices)
        >>> traj = trajs.trajectory_from_index(indices[0])
        >>> pl.plot(traj["particle_time"], traj["particle_position_x"], "-x")
        >>> pl.savefig("orbit")
        """
        mask = np.in1d(self.indices, (index,), assume_unique=True)
        if not np.any(mask):
            print("The particle index %d is not in the list!" % (index))
            raise IndexError
        fields = [field for field in sorted(self.field_data.keys())]
        traj = {}
        traj["particle_time"] = self.times
        traj["particle_index"] = index
        for field in fields:
            traj[field] = self[field][mask,:][0]
        return traj

    @parallel_root_only
    def write_out(self, filename_base):
        """
        Write out particle trajectories to tab-separated ASCII files (one
        for each trajectory) with the field names in the file header. Each
        file is named with a basename and the index number.

        Parameters
        ----------
        filename_base : string
            The prefix for the outputted ASCII files.

        Examples
        --------
        >>> trajs = ParticleTrajectories(my_fns, indices)
        >>> trajs.write_out("orbit_trajectory")
        """
        fields = [field for field in sorted(self.field_data.keys())]
        num_fields = len(fields)
        first_str = "# particle_time\t" + "\t".join(fields)+"\n"
        template_str = "%g\t"*num_fields+"%g\n"
        for ix in range(self.num_indices):
            outlines = [first_str]
            for it in range(self.num_steps):
                outlines.append(template_str %
                                tuple([self.times[it]]+[self[field][ix,it] for field in fields]))
            fid = open(filename_base + "_%d.dat" % self.indices[ix], "w")
            fid.writelines(outlines)
            fid.close()
            del fid

    @parallel_root_only
    def write_out_h5(self, filename):
        """
        Write out all the particle trajectories to a single HDF5 file
        that contains the indices, the times, and the 2D array for each
        field individually

        Parameters
        ----------

        filename : string
            The output filename for the HDF5 file

        Examples
        --------
        >>> trajs = ParticleTrajectories(my_fns, indices)
        >>> trajs.write_out_h5("orbit_trajectories")
        """
        fid = h5py.File(filename, "w")
        fid.create_dataset("particle_indices", dtype=np.int64,
                           data=self.indices)
        fid.close()
        self.times.write_hdf5(filename, dataset_name="particle_times")
        fields = [field for field in sorted(self.field_data.keys())]
        for field in fields:
            self[field].write_hdf5(filename, dataset_name="%s" % field)



"""
Export/Import of volume rendered images.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------
from yt.utilities.on_demand_imports import _h5py as h5py
import numpy as np

from yt.funcs import mylog

def export_rgba(image, fn, h5=True, fits=False, ):
    """
    This function accepts an *image*, of shape (N,M,4) corresponding to r,g,b,a,
    and saves to *fn*.  If *h5* is True, then it will save in hdf5 format.  If
    *fits* is True, it will save in fits format.
    """
    if (not h5 and not fits) or (h5 and fits):
        raise ValueError("Choose either HDF5 or FITS format!")
    if h5:
        f = h5py.File('%s.h5'%fn, "w")
        f.create_dataset("R", data=image[:,:,0])
        f.create_dataset("G", data=image[:,:,1])
        f.create_dataset("B", data=image[:,:,2])
        f.create_dataset("A", data=image[:,:,3])
        f.close()
    if fits:
        from yt.visualization.fits_image import FITSImageData
        data = {}
        data["r"] = image[:,:,0]
        data["g"] = image[:,:,1]
        data["b"] = image[:,:,2]
        data["a"] = image[:,:,3]
        fib = FITSImageData(data)
        fib.writeto('%s.fits'%fn,overwrite=True)

def import_rgba(name, h5=True):
    """
    This function will read back in an HDF5 file, as saved by export_rgba, and
    return the frames to the user.  *name* is the name of the file to be read
    in.
    """
    if h5:
        f = h5py.File(name, "r")
        r = f['R'].value
        g = f['G'].value
        b = f['B'].value
        a = f['A'].value
        f.close()
    else:
        mylog.error('No support for fits import.')
    return np.array([r,g,b,a]).swapaxes(0,2).swapaxes(0,1)

def plot_channel(image, name, cmap='gist_heat', log=True, dex=3, zero_factor=1.0e-10, 
                 label=None, label_color='w', label_size='large'):
    """
    This function will plot a single channel. *image* is an array shaped like
    (N,M), *name* is the pefix for the output filename.  *cmap* is the name of
    the colormap to apply, *log* is whether or not the channel should be
    logged.  Additionally, you may optionally specify the minimum-value cutoff
    for scaling as *dex*, which is taken with respect to the minimum value of
    the image.  *zero_factor* applies a minimum value to all zero-valued
    elements.  Optionally, *label*, *label_color* and *label_size* may be
    specified.
    """
    import matplotlib
    import pylab
    Nvec = image.shape[0]
    image[np.isnan(image)] = 0.0
    ma = image[image>0.0].max()
    image[image==0.0] = ma*zero_factor
    if log:
        mynorm = matplotlib.colors.LogNorm(ma/(10.**dex), ma)

    pylab.clf()
    pylab.gcf().set_dpi(100)
    pylab.gcf().set_size_inches((Nvec/100.0, Nvec/100.0))
    pylab.gcf().subplots_adjust(left=0.0, right=1.0, bottom=0.0, top=1.0, wspace=0.0, hspace=0.0)
    mycm = pylab.cm.get_cmap(cmap)
    if log:
        pylab.imshow(image,cmap=mycm, norm=mynorm, interpolation='nearest')
    else:
        pylab.imshow(image,cmap=mycm, interpolation='nearest')
    if label is not None:
        pylab.text(20, 20,label, color = label_color, size=label_size) 
    pylab.savefig("%s_%s.png" % (name,cmap))
    pylab.clf()

def plot_rgb(image, name, label=None, label_color='w', label_size='large'):
    """
    This will plot the r,g,b channels of an *image* of shape (N,M,3) or
    (N,M,4).  *name* is the prefix of the file name, which will be supplemented
    with "_rgb.png."  *label*, *label_color* and *label_size* may also be
    specified.
    """
    import pylab
    Nvec = image.shape[0]
    image[np.isnan(image)] = 0.0
    if image.shape[2] >= 4:
        image = image[:,:,:3]
    pylab.clf()
    pylab.gcf().set_dpi(100)
    pylab.gcf().set_size_inches((Nvec/100.0, Nvec/100.0))
    pylab.gcf().subplots_adjust(left=0.0, right=1.0, bottom=0.0, top=1.0, wspace=0.0, hspace=0.0)
    pylab.imshow(image, interpolation='nearest')
    if label is not None:
        pylab.text(20, 20, label, color = label_color, size=label_size) 
    pylab.savefig("%s_rgb.png" % name)
    pylab.clf()



from nose.tools import assert_raises
from numpy.testing import assert_equal

from yt import YTQuantity
from yt.testing import fake_random_ds


def test_bad_disk_input():
    # Fixes 1768
    ds = fake_random_ds(16)

    # Test invalid 3d array
    with assert_raises(TypeError) as ex:
        ds.disk(ds.domain_center, [0, 0, 1, 1], (10, 'kpc'), (20, 'kpc'))
    desired = ("Expected an array of size (3,), received 'list' of length 4")
    assert_equal(str(ex.exception), desired)

    # Test invalid float
    with assert_raises(TypeError) as ex:
        ds.disk(ds.domain_center, [0, 0, 1], ds.domain_center, (20, 'kpc'))
    desired = ("Expected a numeric value (or size-1 array),"
               " received 'yt.units.yt_array.YTArray' of length 3")
    assert_equal(str(ex.exception), desired)

    # Test invalid float
    with assert_raises(TypeError) as ex:
        ds.disk(ds.domain_center, [0, 0, 1], (10, 10), (20, 'kpc'))
    desired = ("Expected a numeric value (or tuple of format (float, String)),"
               " received an inconsistent tuple '(10, 10)'.")
    assert_equal(str(ex.exception), desired)

    # Test invalid iterable
    with assert_raises(TypeError) as ex:
        ds.disk(ds.domain_center, [0, 0, 1], (10, 'kpc'),
                (20, 'kpc'), fields=YTQuantity(1, 'kpc'))
    desired = ("Expected an iterable object, received"
               " 'yt.units.yt_array.YTQuantity'")
    assert_equal(str(ex.exception), desired)

    # Test invalid object
    with assert_raises(TypeError) as ex:
        ds.disk(ds.domain_center, [0, 0, 1], (10, 'kpc'),
                (20, 'kpc'), ds=ds.all_data())
    desired = ("Expected an object of 'yt.data_objects.static_output.Dataset' "
               "type, received "
               "'yt.data_objects.selection_data_containers.YTRegion'")
    assert_equal(str(ex.exception), desired)

    # Test valid disk
    ds.disk(ds.domain_center, [0, 0, 1], (10, 'kpc'), (20, 'kpc'))
    ds.disk(ds.domain_center, [0, 0, 1], 10, (20, 'kpc'))



"""
Test symbolic unit handling.




"""
from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from numpy.testing import \
    assert_array_almost_equal_nulp, \
    assert_raises, assert_equal
from nose.tools import assert_true
import operator
from sympy import Symbol
from yt.testing import \
    fake_random_ds, assert_allclose_units, \
    assert_almost_equal
from yt.units.unit_registry import UnitRegistry
from yt.units import electrostatic_unit, elementary_charge
from yt.units.unit_object import default_unit_registry

# dimensions
from yt.units.dimensions import \
    mass, length, time, temperature, energy, magnetic_field, power, rate
# functions
from yt.units.unit_object import get_conversion_factor
# classes
from yt.units.unit_object import Unit, UnitParseError, InvalidUnitOperation
# objects
from yt.units.unit_lookup_table import \
    default_unit_symbol_lut, unit_prefixes, prefixable_units
import yt.units.unit_symbols as unit_symbols
# unit definitions
from yt.utilities.physical_ratios import \
    cm_per_pc, sec_per_year, cm_per_km, cm_per_mpc, \
    mass_sun_grams

def test_no_conflicting_symbols():
    """
    Check unit symbol definitions for conflicts.

    """
    full_set = set(default_unit_symbol_lut.keys())

    # go through all possible prefix combos
    for symbol in default_unit_symbol_lut.keys():
        if symbol in prefixable_units:
            keys = unit_prefixes.keys()
        else:
            keys = [symbol]
        for prefix in keys:
            new_symbol = "%s%s" % (prefix, symbol)

            # test if we have seen this symbol
            if new_symbol in full_set:
                print("Duplicate symbol: %s" % new_symbol)
                raise RuntimeError

            full_set.add(new_symbol)

def test_dimensionless():
    """
    Create dimensionless unit and check attributes.

    """
    u1 = Unit()

    assert_true(u1.is_dimensionless)
    assert_true(u1.expr == 1)
    assert_true(u1.base_value == 1)
    assert_true(u1.dimensions == 1)

    u2 = Unit("")

    assert_true(u2.is_dimensionless)
    assert_true(u2.expr == 1)
    assert_true(u2.base_value == 1)
    assert_true(u2.dimensions == 1)

    assert_equal(u1.latex_repr, '')
    assert_equal(u2.latex_repr, '')

#
# Start init tests
#

def test_create_from_string():
    """
    Create units with strings and check attributes.

    """

    u1 = Unit("g * cm**2 * s**-2")
    assert_true(u1.dimensions == energy)
    assert_true(u1.base_value == 1.0)

    # make sure order doesn't matter
    u2 = Unit("cm**2 * s**-2 * g")
    assert_true(u2.dimensions == energy)
    assert_true(u2.base_value == 1.0)

    # Test rationals
    u3 = Unit("g**0.5 * cm**-0.5 * s**-1")
    assert_true(u3.dimensions == magnetic_field)
    assert_true(u3.base_value == 1.0)

    # sqrt functions
    u4 = Unit("sqrt(g)/sqrt(cm)/s")
    assert_true(u4.dimensions == magnetic_field)
    assert_true(u4.base_value == 1.0)

    # commutative sqrt function
    u5 = Unit("sqrt(g/cm)/s")
    assert_true(u5.dimensions == magnetic_field)
    assert_true(u5.base_value == 1.0)

    # nonzero CGS conversion factor
    u6 = Unit("Msun/pc**3")
    assert_true(u6.dimensions == mass/length**3)
    assert_array_almost_equal_nulp(np.array([u6.base_value]), np.array([mass_sun_grams/cm_per_pc**3]))

    assert_raises(UnitParseError, Unit, 'm**m')
    assert_raises(UnitParseError, Unit, 'm**g')
    assert_raises(UnitParseError, Unit, 'm+g')
    assert_raises(UnitParseError, Unit, 'm-g')


def test_create_from_expr():
    """
    Create units from sympy Exprs and check attributes.

    """
    pc_cgs = cm_per_pc
    yr_cgs = sec_per_year

    # Symbol expr
    s1 = Symbol("pc", positive=True)
    s2 = Symbol("yr", positive=True)
    # Mul expr
    s3 = s1 * s2
    # Pow expr
    s4  = s1**2 * s2**(-1)

    u1 = Unit(s1)
    u2 = Unit(s2)
    u3 = Unit(s3)
    u4 = Unit(s4)

    assert_true(u1.expr == s1)
    assert_true(u2.expr == s2)
    assert_true(u3.expr == s3)
    assert_true(u4.expr == s4)

    assert_allclose_units(u1.base_value, pc_cgs, 1e-12)
    assert_allclose_units(u2.base_value, yr_cgs, 1e-12)
    assert_allclose_units(u3.base_value, pc_cgs * yr_cgs, 1e-12)
    assert_allclose_units(u4.base_value, pc_cgs**2 / yr_cgs, 1e-12)

    assert_true(u1.dimensions == length)
    assert_true(u2.dimensions == time)
    assert_true(u3.dimensions == length * time)
    assert_true(u4.dimensions == length**2 / time)


def test_create_with_duplicate_dimensions():
    """
    Create units with overlapping dimensions. Ex: km/Mpc.

    """

    u1 = Unit("erg * s**-1")
    u2 = Unit("km/s/Mpc")
    km_cgs = cm_per_km
    Mpc_cgs = cm_per_mpc

    assert_true(u1.base_value == 1)
    assert_true(u1.dimensions == power)

    assert_allclose_units(u2.base_value, km_cgs / Mpc_cgs, 1e-12)
    assert_true(u2.dimensions == rate)

def test_create_new_symbol():
    """
    Create unit with unknown symbol.

    """
    u1 = Unit("abc", base_value=42, dimensions=(mass/time))

    assert_true(u1.expr == Symbol("abc", positive=True))
    assert_true(u1.base_value == 42)
    assert_true(u1.dimensions == mass / time)

    u1 = Unit("abc", base_value=42, dimensions=length**3)

    assert_true(u1.expr == Symbol("abc", positive=True))
    assert_true(u1.base_value == 42)
    assert_true(u1.dimensions == length**3)

    u1 = Unit("abc", base_value=42, dimensions=length*(mass*length))

    assert_true(u1.expr == Symbol("abc", positive=True))
    assert_true(u1.base_value == 42)
    assert_true( u1.dimensions == length**2*mass)

    assert_raises(UnitParseError, Unit, 'abc', base_value=42,
                  dimensions=length**length)
    assert_raises(UnitParseError, Unit, 'abc', base_value=42,
                  dimensions=length**(length*length))
    assert_raises(UnitParseError, Unit, 'abc', base_value=42,
                  dimensions=length-mass)
    assert_raises(UnitParseError, Unit, 'abc', base_value=42,
                  dimensions=length+mass)

def test_create_fail_on_unknown_symbol():
    """
    Fail to create unit with unknown symbol, without base_value and dimensions.

    """
    try:
        Unit(Symbol("jigawatts"))
    except UnitParseError:
        assert_true(True)
    else:
        assert_true(False)

def test_create_fail_on_bad_symbol_type():
    """
    Fail to create unit with bad symbol type.

    """
    try:
        Unit([1])  # something other than Expr and str
    except UnitParseError:
        assert_true(True)
    else:
        assert_true(False)

def test_create_fail_on_bad_dimensions_type():
    """
    Fail to create unit with bad dimensions type.

    """
    try:
        Unit("a", base_value=1, dimensions="(mass)")
    except UnitParseError:
        assert_true(True)
    else:
        assert_true(False)


def test_create_fail_on_dimensions_content():
    """
    Fail to create unit with bad dimensions expr.

    """
    a = Symbol("a")

    try:
        Unit("a", base_value=1, dimensions=a)
    except UnitParseError:
        pass
    else:
        assert_true(False)


def test_create_fail_on_base_value_type():
    """
    Fail to create unit with bad base_value type.

    """
    try:
        Unit("a", base_value="a", dimensions=(mass/time))
    except UnitParseError:
        assert_true(True)
    else:
        assert_true(False)

#
# End init tests
#

def test_string_representation():
    """
    Check unit string representation.

    """
    pc = Unit("pc")
    Myr = Unit("Myr")
    speed = pc / Myr
    dimensionless = Unit()

    assert_true(str(pc) == "pc")
    assert_true(str(Myr) == "Myr")
    assert_true(str(speed) == "pc/Myr")
    assert_true(repr(speed) == "pc/Myr")
    assert_true(str(dimensionless) == "dimensionless")

#
# Start operation tests
#

def test_multiplication():
    """
    Multiply two units.

    """
    msun_cgs = mass_sun_grams
    pc_cgs = cm_per_pc

    # Create symbols
    msun_sym = Symbol("Msun", positive=True)
    pc_sym = Symbol("pc", positive=True)
    s_sym = Symbol("s", positive=True)

    # Create units
    u1 = Unit("Msun")
    u2 = Unit("pc")

    # Mul operation
    u3 = u1 * u2

    assert_true(u3.expr == msun_sym * pc_sym)
    assert_allclose_units(u3.base_value, msun_cgs * pc_cgs, 1e-12)
    assert_true(u3.dimensions == mass * length)

    # Pow and Mul operations
    u4 = Unit("pc**2")
    u5 = Unit("Msun * s")

    u6 = u4 * u5

    assert_true(u6.expr == pc_sym**2 * msun_sym * s_sym)
    assert_allclose_units(u6.base_value, pc_cgs**2 * msun_cgs, 1e-12)
    assert_true(u6.dimensions == length**2 * mass * time)


def test_division():
    """
    Divide two units.

    """
    pc_cgs = cm_per_pc
    km_cgs = cm_per_km

    # Create symbols
    pc_sym = Symbol("pc", positive=True)
    km_sym = Symbol("km", positive=True)
    s_sym = Symbol("s", positive=True)

    # Create units
    u1 = Unit("pc")
    u2 = Unit("km * s")

    u3 = u1 / u2

    assert_true(u3.expr == pc_sym / (km_sym * s_sym))
    assert_allclose_units(u3.base_value, pc_cgs / km_cgs, 1e-12)
    assert_true(u3.dimensions == 1 / time)


def test_power():
    """
    Take units to some power.

    """
    from sympy import nsimplify

    pc_cgs = cm_per_pc
    mK_cgs = 1e-3
    u1_dims = mass * length**2 * time**-3 * temperature**4
    u1 = Unit("g * pc**2 * s**-3 * mK**4")

    u2 = u1**2

    assert_true(u2.dimensions == u1_dims**2)
    assert_allclose_units(u2.base_value, (pc_cgs**2 * mK_cgs**4)**2, 1e-12)

    u3 = u1**(-1.0/3)

    assert_true(u3.dimensions == nsimplify(u1_dims**(-1.0/3)))
    assert_allclose_units(u3.base_value, (pc_cgs**2 * mK_cgs**4)**(-1.0/3), 1e-12)


def test_equality():
    """
    Check unit equality with different symbols, but same dimensions and base_value.

    """
    u1 = Unit("km * s**-1")
    u2 = Unit("m * ms**-1")

    assert_true(u1 == u2)

#
# End operation tests.
#

def test_base_equivalent():
    """
    Check base equivalent of a unit.

    """
    Msun_cgs = mass_sun_grams
    Mpc_cgs = cm_per_mpc

    u1 = Unit("Msun * Mpc**-3")
    u2 = Unit("g * cm**-3")
    u3 = u1.get_base_equivalent()

    assert_true(u2.expr == u3.expr)
    assert_true(u2 == u3)

    assert_allclose_units(u1.base_value, Msun_cgs / Mpc_cgs**3, 1e-12)
    assert_true(u2.base_value == 1)
    assert_true(u3.base_value == 1)

    mass_density = mass / length**3

    assert_true(u1.dimensions == mass_density)
    assert_true(u2.dimensions == mass_density)
    assert_true(u3.dimensions == mass_density)

    assert_allclose_units(get_conversion_factor(u1, u3)[0], Msun_cgs / Mpc_cgs**3, 1e-12)

def test_is_code_unit():
    ds = fake_random_ds(64, nprocs=1)
    u1 = Unit('code_mass', registry=ds.unit_registry)
    u2 = Unit('code_mass/code_length', registry=ds.unit_registry)
    u3 = Unit('code_velocity*code_mass**2', registry=ds.unit_registry)
    u4 = Unit('code_time*code_mass**0.5', registry=ds.unit_registry)
    u5 = Unit('code_mass*g', registry=ds.unit_registry)
    u6 = Unit('g/cm**3')

    assert_true(u1.is_code_unit)
    assert_true(u2.is_code_unit)
    assert_true(u3.is_code_unit)
    assert_true(u4.is_code_unit)
    assert_true(not u5.is_code_unit)
    assert_true(not u6.is_code_unit)

def test_temperature_offsets():
    u1 = Unit('degC')
    u2 = Unit('degF')

    assert_raises(InvalidUnitOperation, operator.mul, u1, u2)
    assert_raises(InvalidUnitOperation, operator.truediv, u1, u2)

def test_latex_repr():
    ds = fake_random_ds(64, nprocs=1)

    # create a fake comoving unit
    ds.unit_registry.add('pccm', ds.unit_registry.lut['pc'][0]/(1+2), length,
                         "\\rm{pc}/(1+z)")

    test_unit = Unit('Mpccm', registry=ds.unit_registry)
    assert_almost_equal(test_unit.base_value, cm_per_mpc/3)
    assert_equal(test_unit.latex_repr, r'\rm{Mpc}/(1+z)')

    test_unit = Unit('code_mass', registry=ds.unit_registry)
    assert_equal(test_unit.latex_repr, '\\rm{code\\ mass}')

    test_unit = Unit('code_mass/code_length**3', registry=ds.unit_registry)
    assert_equal(test_unit.latex_repr,
                 '\\frac{\\rm{code\\ mass}}{\\rm{code\\ length}^{3}}')

    test_unit = Unit('cm**-3', base_value=1.0, registry=ds.unit_registry)
    assert_equal(test_unit.latex_repr, '\\frac{1}{\\rm{cm}^{3}}')

    test_unit = Unit('m_geom/l_geom**3')
    assert_equal(test_unit.latex_repr, '\\frac{1}{M_\\odot^{2}}')

    test_unit = Unit('1e9*cm')
    assert_equal(test_unit.latex_repr, '1.0 \\times 10^{9}\\ \\rm{cm}')

def test_latitude_longitude():
    lat = unit_symbols.lat
    lon = unit_symbols.lon
    deg = unit_symbols.deg
    assert_equal(lat.units.base_offset, 90.0)
    assert_equal((deg*90.0).in_units("lat").value, 0.0)
    assert_equal((deg*180).in_units("lat").value, -90.0)
    assert_equal((lat*0.0).in_units("deg"), deg*90.0)
    assert_equal((lat*-90).in_units("deg"), deg*180)

    assert_equal(lon.units.base_offset, -180.0)
    assert_equal((deg*0.0).in_units("lon").value, -180.0)
    assert_equal((deg*90.0).in_units("lon").value, -90.0)
    assert_equal((deg*180).in_units("lon").value, 0.0)
    assert_equal((deg*360).in_units("lon").value, 180.0)

    assert_equal((lon*-180.0).in_units("deg"), deg*0.0)
    assert_equal((lon*-90.0).in_units("deg"), deg*90.0)
    assert_equal((lon*0.0).in_units("deg"), deg*180.0)
    assert_equal((lon*180.0).in_units("deg"), deg*360)

def test_registry_json():
    reg = UnitRegistry()
    json_reg = reg.to_json()
    unserialized_reg = UnitRegistry.from_json(json_reg)

    assert_equal(reg.lut, unserialized_reg.lut)

def test_creation_from_ytarray():
    u1 = Unit(electrostatic_unit)
    assert_equal(str(u1), 'esu')
    assert_equal(u1, Unit('esu'))
    assert_equal(u1, electrostatic_unit.units)

    u2 = Unit(elementary_charge)
    assert_equal(str(u2), '4.8032056e-10*esu')
    assert_equal(u2, Unit('4.8032056e-10*esu'))
    assert_equal(u1, elementary_charge.units)

    assert_equal((u1/u2).base_value, electrostatic_unit/elementary_charge)

    assert_raises(UnitParseError, Unit, [1, 2, 3]*elementary_charge)

def test_list_same_dimensions():
    reg = default_unit_registry
    for name1, u1 in reg.unit_objs.items():
        for name2 in reg.list_same_dimensions(u1):
            if name2 == name1: continue
            if name2 in reg.unit_objs:
                dim2 = reg.unit_objs[name2].dimensions
            else:
                _, dim2, _, _ = reg.lut[name2]
            assert_true(u1.dimensions is dim2)



"""
Title:   utils.py
Purpose: Contains utility functions for yt answer tests
Notes:
"""
import functools
import hashlib
import inspect
import os

import numpy as np
import pytest
import yaml

from yt.config import ytcfg
from yt.convenience import load, simulation
from yt.data_objects.selection_data_containers import YTRegion
from yt.data_objects.static_output import Dataset
from yt.frontends.ytdata.api import save_as_dataset
from yt.units.yt_array import \
    YTArray, \
    YTQuantity
from yt.utilities.exceptions import \
    YTOutputNotIdentified
import yt.visualization.particle_plots as particle_plots
import yt.visualization.plot_window as pw
import yt.visualization.profile_plotter as profile_plotter
from yt.visualization.volume_rendering.scene import Scene


def _streamline_for_io(params):
    r"""
    Put test results in a more io-friendly format.

    Many of yt's tests use objects such as tuples as test parameters
    (fields, for instance), but when these objects are written to a
    yaml file, yaml includes python specific anchors that make the file
    harder to read and less portable. The goal of this function is to
    convert these objects to strings (using __repr__() has it's own
    issues) in order to solve this problem.

    Parameters:
    -----------
        params : dict
            The dictionary of test parameters in the form
            {param_name : param_value}.

    Returns:
    --------
        streamlined_params : dict
            The dictionary of parsed and converted
            {param_name : param_value} pairs.
    """
    streamlined_params = {}
    for key, value in params.items():
        # Check for user-defined functions
        if inspect.isfunction(key):
            key = key.__name__
        if inspect.isfunction(value):
            value = value.__name__
        # The key can be nested iterables, e.g., 
        # d = [None, ('sphere', (center, (0.1, 'unitary')))] so we need
        # to use recursion
        if not isinstance(key, str) and hasattr(key, '__iter__'):
            key = _iterable_to_string(key)
        # The value can also be nested iterables
        if not isinstance(value, str) and hasattr(value, '__iter__'):
            value = _iterable_to_string(value)
        # Scene objects need special treatment to make them more IO friendly
        if isinstance(value, Scene):
            value = 'Scene' 
        elif isinstance(value, YTRegion):
            value = 'Region'
        streamlined_params[key] = value
    return streamlined_params

def _iterable_to_string(iterable):
    r"""
    An extension of streamline_for_io that does the work of making an
    iterable more io-friendly.

    Parameters:
    -----------
        iterable : python iterable
            The object to be parsed and converted.

    Returns:
    --------
        result : str
            The io-friendly version of the passed iterable.
    """
    result = iterable.__class__.__name__ 
    for elem in iterable:
        # Check for user-defined functions
        if inspect.isfunction(elem):
            result += '_' + elem.__name__
        # Non-string iterables (e.g., lists, tuples, etc.)
        elif not isinstance(elem, str) and hasattr(elem, '__iter__'):
            result += '_' + _iterable_to_string(elem)
        # Non-string non-iterables (ints, floats, etc.)
        elif not isinstance(elem, str) and not hasattr(elem, '__iter__'):
            result += '_' + str(elem)
        # Strings
        elif isinstance(elem, str):
            result += '_' + elem
    return result

def _hash_results(results):
    r"""
    Driver function for hashing the test result.

    Parameters:
    -----------
        results : dict
            Dictionary of {test_name : test_result} pairs.

    Returns:
    --------
        results : dict
            Same as the passed results, but the test_results are now
            hex digests of the hashed test_result.
    """
    # Here, results should be comprised of only the tests, not the test
    # parameters
    for test_name, test_value in results.items():
        # These tests have issues with python-specific anchors and so
        # are already hashed
        # (see their definitions in yt/utilites/answer_testing/answer_tests.py)
        if test_name in ['projection_values', 'pixelized_projection_values', 'grid_values']:
            continue
        else:
            results[test_name] = generate_hash(test_value)
    return results

def _hash_dict(data):
    r"""
    Specifically handles hashing a dictionary object. 

    Parameters:
    -----------
        data : dict
            The dictionary to be hashed.

    Returns:
    --------
        hd.hexdigest : str
            The hex digest of the hashed dictionary.
    """
    hd = None
    for key, value in sorted(data.items()):
        if hd is None:
            hd = hashlib.md5(bytearray(key.encode('utf8')) + bytearray(value))
        else:
            hd.update(bytearray(key.encode('utf8')) + bytearray(value))
    return hd.hexdigest()

def generate_hash(data):
    r"""
    Actually performs the hash operation. 

    Parameters:
    -----------
        data : python object 
            Data to be hashed.

    Returns:
    --------
        hd : str
            Hex digest of the hashed data.
    """
    # Sometimes md5 complains that the data is not contiguous, so we
    # make it so here
    if isinstance(data, np.ndarray):
        data = np.ascontiguousarray(data)
    # Try to hash. Some tests return hashable types (like ndarrays) and
    # others don't (such as dictionaries)
    try:
        hd = hashlib.md5(data).hexdigest()
    # Handle those tests that return non-hashable types. This is done
    # here instead of in the tests themselves to try and reduce boilerplate
    # and provide a central location where all of this is done in case it needs
    # to be changed
    except TypeError:
        if isinstance(data, dict):
            hd = _hash_dict(data)
        else:
            raise TypeError
    return hd

def _save_result(data, outputFile):
    r"""
    Saves the test results to the desired answer file. 

    Parameters:
    -----------
        data : dict
            Test results to be saved.

        outputFile : str
            Name of file to save results to.

    Returns:
    --------
        None 
    """
    with open(outputFile, 'a') as f:
        yaml.dump(data, f)

def _compare_result(data, outputFile):
    r"""
    Compares the just-generated test results to those that are already
    saved.

    Parameters:
    -----------
        data : dict
            Just-generated test results.

        outputFile : str
            Name of file where answers are already saved.
    """
    # Load the saved data
    with open(outputFile, 'r') as f:
        savedData = yaml.safe_load(f)
    # Define the comparison function
    def _check_vals(newVals, oldVals):
        for key, value in newVals.items():
            if isinstance(value, dict):
                _check_vals(value, oldVals[key])
            else:
                assert value == oldVals[key]
    # Compare
    _check_vals(data, savedData)

def _handle_hashes(save_dir_name, fname, hashes, answer_store):
    r"""
    Driver function for deciding whether to save the test results or
    compare them to already saved results.

    Parameters:
    -----------
        save_dir_name : str
            Name of the directory to save results or where results are
            already saved.

        fname : str
            Name of the file to either save results to or where results
            are already saved.

        hashes : dict
            The just-generated test results.

        answer_store : bool
            If true, save the just-generated test results, otherwise,
            compare them to the previously saved results.
    """
    # Set up the answer file in the answer directory
    answer_file = os.path.join(save_dir_name, fname)
    # Save the result
    if answer_store:
        _save_result(hashes, answer_file)
    # Compare to already saved results
    else:
        _compare_result(hashes, answer_file)


def _save_arrays(save_dir_name, fbasename, arrays, answer_store):
    r"""
    Driver routine for either saving the raw arrays resulting from the
    tests, or compare them to previously saved results.

    Parameters:
    -----------
        save_dir_name : str
            Name of the directory to save results or where results are
            already saved.

        fbasename : str
            Base name (no extension) of the file to either save results
            to or where results are already saved.

        arrays : dict
            The raw arrays generated from the tests, with the test name
            as a key.

        answer_store : bool
            If true, save the just-generated test results, otherwise,
            compare them to the previously saved results.
    """
    pass


def can_run_ds(ds_fn, file_check = False):
    r"""
    Validates whether or not a given input can be loaded and used as a
    Dataset object.
    """
    if isinstance(ds_fn, Dataset):
        return True
    path = ytcfg.get("yt", "test_data_dir")
    if not os.path.isdir(path):
        return False
    if file_check:
        return os.path.isfile(os.path.join(path, ds_fn))
    try:
        load(ds_fn)
        return True
    except YTOutputNotIdentified:
        return False


def can_run_sim(sim_fn, sim_type, file_check = False):
    r"""
    Validates whether or not a given input can be used as a simulation
    time-series object.
    """
    path = ytcfg.get("yt", "test_data_dir")
    if not os.path.isdir(path):
        return False
    if file_check:
        return os.path.isfile(os.path.join(path, sim_fn))
    try:
        simulation(sim_fn, sim_type)
    except YTOutputNotIdentified:
        return False
    return True 


def data_dir_load(ds_fn, cls = None, args = None, kwargs = None):
    r"""
    Loads a sample dataset from the designated test_data_dir for use in
    testing.
    """
    args = args or ()
    kwargs = kwargs or {}
    path = ytcfg.get("yt", "test_data_dir")
    if isinstance(ds_fn, Dataset): return ds_fn
    if not os.path.isdir(path):
        return False
    if cls is None:
        ds = load(ds_fn, *args, **kwargs)
    else:
        ds = cls(os.path.join(path, ds_fn), *args, **kwargs)
    ds.index
    return ds


def requires_ds(ds_fn, file_check = False):
    r"""
    Meta-wrapper for specifying required data for a test and
    checking if said data exists.
    """
    def ffalse(func):
        @functools.wraps(func)
        def skip(*args, **kwargs):
            msg = "{} not found, skipping {}.".format(ds_fn, func.__name__)
            pytest.fail(msg)
        return skip
    def ftrue(func):
        @functools.wraps(func)
        def true_wrapper(*args, **kwargs):
            return func
        return true_wrapper
    if not can_run_ds(ds_fn, file_check):
        return ffalse
    else:
        return ftrue


def requires_sim(sim_fn, sim_type, file_check = False):
    r"""
    Meta-wrapper for specifying a required simulation for a test and
    checking if said simulation exists.
    """
    def ffalse(func):
        @functools.wraps(func)
        def skip(*args, **kwargs):
            msg = "{} not found, skipping {}.".format(sim_fn, func.__name__)
            pytest.fail(msg)
        return skip
    def ftrue(func):
        @functools.wraps(func)
        def true_wrapper(*args, **kwargs):
            return func
        return true_wrapper
    if not can_run_sim(sim_fn, sim_type, file_check):
        return ffalse
    else:
        return ftrue


def create_obj(ds, obj_type):
    # obj_type should be tuple of
    #  ( obj_name, ( args ) )
    if obj_type is None:
        return ds.all_data()
    cls = getattr(ds, obj_type[0])
    obj = cls(*obj_type[1])
    return obj


def compare_unit_attributes(ds1, ds2):
    r"""
    Checks to make sure that the length, mass, time, velocity, and
    magnetic units are the same for two different dataset objects.
    """
    attrs = ('length_unit', 'mass_unit', 'time_unit',
             'velocity_unit', 'magnetic_unit')
    for attr in attrs:
        u1 = getattr(ds1, attr, None)
        u2 = getattr(ds2, attr, None)
        assert u1 == u2


def fake_halo_catalog(data):
    filename = "catalog.0.h5"
    ftypes = dict((field, '.') for field in data)
    extra_attrs = {"data_type": "halo_catalog",
                   "num_halos": data['particle_mass'].size}
    ds = {'cosmological_simulation': 1,
          'omega_lambda': 0.7,
          'omega_matter': 0.3,
          'hubble_constant': 0.7,
          'current_redshift': 0,
          'current_time': YTQuantity(1, 'yr'),
          'domain_left_edge': YTArray(np.zeros(3), 'cm'),
          'domain_right_edge': YTArray(np.ones(3), 'cm')}
    save_as_dataset(ds, filename, data, field_types=ftypes,
        extra_attrs=extra_attrs
    )
    return filename


def _create_plot_window_attribute_plot(ds, ptype, field, axis, pkwargs = None):
    r"""
    Convenience function used in plot_window_attribute_test.

    Parameters:
    -----------
        ds : Dataset
            The Dataset object from which the plotting data is taken.

        ptype : string
            Type of plot to make (e.g., SlicePlot).

        field : yt field
            The field (e.g, density) to plot.

        axis : int
            The plot axis to plot or project along.
            
        pkwargs : dict
            Any keywords to be passed when creating the plot.
    """
    if ptype is None:
        raise RuntimeError('Must explicitly request a plot type')
    cls = getattr(pw, ptype, None)
    if cls is None:
        cls = getattr(particle_plots, ptype)
    plot = cls(*(ds, axis, field), **pkwargs)
    return plot


def _create_phase_plot_attribute_plot(data_source, x_field, y_field, z_field,
                plot_type, plot_kwargs=None):
    r"""
    Convenience function used in phase_plot_attribute_test.

    Parameters:
    -----------
        data_source : Dataset object
            The Dataset object from which the plotting data is taken.
        
        x_field : yt field
            Field to plot on x-axis.
        
        y_field : yt field
            Field to plot on y-axis.

        z_field : yt field
            Field to plot on z-axis.

        plot_type : string
            Type of plot to make (e.g., SlicePlot).
            
        plot_kwargs : dict
            Any keywords to be passed when creating the plot.
    """
    if plot_type is None:
        raise RuntimeError('Must explicitly request a plot type')
    cls = getattr(profile_plotter, plot_type, None)
    if cls is None:
        cls = getattr(particle_plots, plot_type)
    plot = cls(*(data_source, x_field, y_field, z_field), **plot_kwargs)
    return plot



from yt.testing import fake_random_ds
from yt.utilities.answer_testing.level_sets_tests import ExtractConnectedSetsTest

def test_connected_sets():
    ds = fake_random_ds(16, nprocs=8, particles=16**3)
    data_source = ds.disk([0.5, 0.5, 0.5], [0., 0., 1.], (8, 'kpc'), (1, 'kpc'))
    field = ("gas", "density")
    min_val, max_val = data_source[field].min()/2, data_source[field].max()/2
    data_source.extract_connected_sets(field, 5, min_val, max_val,
                                       log_space=True, cumulative=True)
    yield ExtractConnectedSetsTest(ds, data_source, field, 5, min_val, max_val)



from yt.testing import \
    fake_random_ds, \
    assert_equal, \
    assert_true
from yt.units.yt_array import \
    uconcatenate

def _get_dobjs(c):
    dobjs = [("sphere", ("center", (1.0, "unitary"))),
             ("sphere", ("center", (0.1, "unitary"))),
             ("ortho_ray", (0, (c[1], c[2]))),
             ("slice", (0, c[0])),
             #("disk", ("center", [0.1, 0.3, 0.6],
             #           (0.2, 'unitary'), (0.1, 'unitary'))),
             ("cutting", ([0.1, 0.3, 0.6], 'center')),
             ("all_data", ()),
            ]
    return dobjs

def test_chunking():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(64, nprocs = nprocs)
        c = (ds.domain_right_edge + ds.domain_left_edge)/2.0 
        c += ds.arr(0.5/ds.domain_dimensions, "code_length")
        for dobj in _get_dobjs(c):
            obj = getattr(ds, dobj[0])(*dobj[1])
            coords = {'f':{}, 'i':{}}
            for t in ["io", "all", "spatial"]:
                coords['i'][t] = []
                coords['f'][t] = []
                for chunk in obj.chunks(None, t):
                    coords['f'][t].append(chunk.fcoords[:,:])
                    coords['i'][t].append(chunk.icoords[:,:])
                coords['f'][t] = uconcatenate(coords['f'][t])
                coords['i'][t] = uconcatenate(coords['i'][t])
                coords['f'][t].sort()
                coords['i'][t].sort()
            assert_equal(coords['f']['io'], coords['f']['all'])
            assert_equal(coords['f']['io'], coords['f']['spatial'])
            assert_equal(coords['i']['io'], coords['i']['all'])
            assert_equal(coords['i']['io'], coords['i']['spatial'])

def test_ds_hold():
    ds1 = fake_random_ds(64)
    ds2 = fake_random_ds(128)
    dd = ds1.all_data()
    assert_true(dd.ds.__hash__() == ds1.__hash__()) # dd.ds is a weakref, so can't use "is"
    assert_true(dd.index is ds1.index)
    assert_equal(dd["ones"].size, 64**3)
    with dd._ds_hold(ds2):
        assert_true(dd.ds.__hash__() == ds2.__hash__())
        assert_true(dd.index is ds2.index)
        assert_equal(dd["ones"].size, 128**3)
    assert_true(dd.ds.__hash__() == ds1.__hash__())
    assert_true(dd.index is ds1.index)
    assert_equal(dd["ones"].size, 64**3)



"""
Enzo frontend tests



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.testing import \
    assert_almost_equal, \
    assert_equal, \
    requires_file, \
    units_override_check, \
    assert_array_equal, \
    assert_allclose_units
from yt.utilities.answer_testing.framework import \
    requires_ds, \
    small_patch_amr, \
    big_patch_amr, \
    data_dir_load, \
    AnalyticHaloMassFunctionTest, \
    SimulatedHaloMassFunctionTest
from yt.visualization.plot_window import \
    SlicePlot
from yt.frontends.enzo.api import EnzoDataset
from yt.frontends.enzo.fields import NODAL_FLAGS

_fields = ("temperature", "density", "velocity_magnitude",
           "velocity_divergence")

two_sphere_test = 'ActiveParticleTwoSphere/DD0011/DD0011'
active_particle_cosmology = 'ActiveParticleCosmology/DD0046/DD0046'
ecp = "enzo_cosmology_plus/DD0046/DD0046"
g30 = "IsolatedGalaxy/galaxy0030/galaxy0030"
enzotiny = "enzo_tiny_cosmology/DD0046/DD0046"
toro1d = "ToroShockTube/DD0001/data0001"
kh2d = "EnzoKelvinHelmholtz/DD0011/DD0011"
mhdctot = "MHDCTOrszagTang/DD0004/data0004"
dnz = "DeeplyNestedZoom/DD0025/data0025"
p3mini = "PopIII_mini/DD0034/DD0034"


def color_conservation(ds):
    species_names = ds.field_info.species_names
    dd = ds.all_data()
    dens_yt = dd["density"].copy()
    # Enumerate our species here
    for s in sorted(species_names):
        if s == "El": continue
        dens_yt -= dd["%s_density" % s]
    dens_yt -= dd["metal_density"]
    delta_yt = np.abs(dens_yt / dd["density"])
    # Now we compare color conservation to Enzo's color conservation
    dd = ds.all_data()
    dens_enzo = dd["Density"].copy()
    for f in sorted(ds.field_list):
        ff = f[1]
        if not ff.endswith("_Density"):
            continue
        start_strings = ["Electron_", "SFR_", "Forming_Stellar_",
                         "Dark_Matter", "Star_Particle_"]
        if any([ff.startswith(ss) for ss in start_strings]):
            continue
        dens_enzo -= dd[f]
    delta_enzo = np.abs(dens_enzo / dd["Density"])
    np.testing.assert_almost_equal(delta_yt, delta_enzo)


def check_color_conservation(ds):
    species_names = ds.field_info.species_names
    dd = ds.all_data()
    dens_yt = dd["density"].copy()
    # Enumerate our species here
    for s in sorted(species_names):
        if s == "El": continue
        dens_yt -= dd["%s_density" % s]
    dens_yt -= dd["metal_density"]
    delta_yt = np.abs(dens_yt / dd["density"])

    # Now we compare color conservation to Enzo's color conservation
    dd = ds.all_data()
    dens_enzo = dd["Density"].copy()
    for f in sorted(ds.field_list):
        ff = f[1]
        if not ff.endswith("_Density"):
            continue
        start_strings = ["Electron_", "SFR_", "Forming_Stellar_",
                         "Dark_Matter", "Star_Particle_"]
        if any([ff.startswith(ss) for ss in start_strings]):
            continue
        dens_enzo -= dd[f]
    delta_enzo = np.abs(dens_enzo / dd["Density"])
    return assert_almost_equal, delta_yt, delta_enzo

m7 = "DD0010/moving7_0010"
@requires_ds(m7)
def test_moving7():
    ds = data_dir_load(m7)
    assert_equal(str(ds), "moving7_0010")
    for test in small_patch_amr(m7, _fields):
        test_moving7.__name__ = test.description
        yield test

@requires_ds(g30, big_data=True)
def test_galaxy0030():
    ds = data_dir_load(g30)
    yield check_color_conservation(ds)
    assert_equal(str(ds), "galaxy0030")
    for test in big_patch_amr(ds, _fields):
        test_galaxy0030.__name__ = test.description
        yield test
    assert_equal(ds.particle_type_counts, {'io': 1124453})

@requires_ds(toro1d)
def test_toro1d():
    ds = data_dir_load(toro1d)
    assert_equal(str(ds), 'data0001')
    for test in small_patch_amr(ds, ds.field_list):
        test_toro1d.__name__ = test.description
        yield test

@requires_ds(kh2d)
def test_kh2d():
    ds = data_dir_load(kh2d)
    assert_equal(str(ds), 'DD0011')
    for test in small_patch_amr(ds, ds.field_list):
        test_kh2d.__name__ = test.description
        yield test

@requires_ds(enzotiny)
def test_simulated_halo_mass_function():
    ds = data_dir_load(enzotiny)
    for finder in ["fof", "hop"]:
        yield SimulatedHaloMassFunctionTest(ds, finder)

@requires_ds(enzotiny)
def test_analytic_halo_mass_function():
    ds = data_dir_load(enzotiny)
    for fit in range(1, 6):
        yield AnalyticHaloMassFunctionTest(ds, fit)

@requires_ds(ecp, big_data=True)
def test_ecp():
    ds = data_dir_load(ecp)
    # Now we test our species fields
    yield check_color_conservation(ds)

@requires_file(enzotiny)
def test_units_override():
    units_override_check(enzotiny)

@requires_ds(ecp, big_data=True)
def test_nuclei_density_fields():
    ds = data_dir_load(ecp)
    ad = ds.all_data()
    assert_array_equal(ad["H_nuclei_density"],
                       (ad["H_number_density"] + ad["H_p1_number_density"]))
    assert_array_equal(ad["He_nuclei_density"],
        (ad["He_number_density"] +
         ad["He_p1_number_density"] + ad["He_p2_number_density"]))

@requires_file(enzotiny)
def test_EnzoDataset():
    assert isinstance(data_dir_load(enzotiny), EnzoDataset)

@requires_file(two_sphere_test)
@requires_file(active_particle_cosmology)
def test_active_particle_datasets():
    two_sph = data_dir_load(two_sphere_test)
    assert 'AccretingParticle' in two_sph.particle_types_raw
    assert 'io' not in two_sph.particle_types_raw
    assert 'all' in two_sph.particle_types
    assert_equal(len(two_sph.particle_unions), 1)
    pfields = ['GridID', 'creation_time', 'dynamical_time',
               'identifier', 'level', 'metallicity', 'particle_mass']
    pfields += ['particle_position_%s' % d for d in 'xyz']
    pfields += ['particle_velocity_%s' % d for d in 'xyz']

    acc_part_fields = \
        [('AccretingParticle', pf) for pf in ['AccretionRate'] + pfields]

    real_acc_part_fields = sorted(
        [f for f in two_sph.field_list if f[0] == 'AccretingParticle'])
    assert_equal(acc_part_fields, real_acc_part_fields)


    apcos = data_dir_load(active_particle_cosmology)
    assert_equal(['CenOstriker', 'DarkMatter'], apcos.particle_types_raw)
    assert 'all' in apcos.particle_unions

    apcos_fields = [('CenOstriker', pf) for pf in pfields]

    real_apcos_fields = sorted(
        [f for f in apcos.field_list if f[0] == 'CenOstriker'])

    assert_equal(apcos_fields, real_apcos_fields)

    assert_equal(apcos.particle_type_counts,
                 {'CenOstriker': 899755, 'DarkMatter': 32768})

@requires_file(mhdctot)
def test_face_centered_mhdct_fields():
    ds = data_dir_load(mhdctot)

    ad = ds.all_data()
    grid = ds.index.grids[0]

    for field, flag in NODAL_FLAGS.items():
        dims = ds.domain_dimensions
        assert_equal(ad[field].shape, (dims.prod(), 2*sum(flag)))
        assert_equal(grid[field].shape, tuple(dims) + (2*sum(flag),))

    # Average of face-centered fields should be the same as cell-centered field
    assert (ad['BxF'].sum(axis=-1)/2 == ad['Bx']).all()
    assert (ad['ByF'].sum(axis=-1)/2 == ad['By']).all()
    assert (ad['BzF'].sum(axis=-1)/2 == ad['Bz']).all()

@requires_file(dnz)
def test_deeply_nested_zoom():
    ds = data_dir_load(dnz)

    # carefully chosen to just barely miss a grid in the middle of the image
    center = [0.4915073260199302, 0.5052605316800006, 0.4905805557500548]

    plot = SlicePlot(ds, 'z', 'density', width=(0.001, 'pc'),
                     center=center)

    image = plot.frb['density']

    assert (image > 0).all()

    v, c = ds.find_max('density')

    assert_allclose_units(v, ds.quan(0.005878286377124154, 'g/cm**3'))

    c_actual = [0.49150732540021, 0.505260532936791, 0.49058055816398]
    c_actual = ds.arr(c_actual, 'code_length')
    assert_allclose_units(c, c_actual)

    assert_equal(max([g['density'].max() for g in ds.index.grids]), v)

@requires_file(kh2d)
def test_2d_grid_shape():
    # see issue #1601
    # we want to make sure that accessing data on a grid object
    # returns a 3D array with a dummy dimension.
    ds = data_dir_load(kh2d)
    g = ds.index.grids[1]
    assert g['density'].shape == (128, 100, 1)

@requires_file(p3mini)
def test_nonzero_omega_radiation():
    """
    Test support for non-zero omega_radiation cosmologies.
    """
    ds = data_dir_load(p3mini)

    assert_equal(ds.omega_radiation, ds.cosmology.omega_radiation)

    tratio = ds.current_time / \
      ds.cosmology.t_from_z(ds.current_redshift)
    assert_almost_equal(tratio, 1, 4,
      err_msg='Simulation time not consistent with cosmology calculator.')



"""
Rockstar data-file handling function




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import os

from yt.funcs import mylog

from yt.utilities.exceptions import \
    YTDomainOverflow

from yt.utilities.io_handler import \
    BaseIOHandler

from .definitions import halo_dts
from yt.utilities.lib.geometry_utils import compute_morton

from operator import attrgetter

class IOHandlerRockstarBinary(BaseIOHandler):
    _dataset_type = "rockstar_binary"

    def __init__(self, *args, **kwargs):
        super(IOHandlerRockstarBinary, self).__init__(*args, **kwargs)
        self._halo_dt = halo_dts[self.ds.parameters['format_revision']]

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.
        chunks = list(chunks)
        data_files = set([])
        # Only support halo reading for now.
        assert(len(ptf) == 1)
        assert(list(ptf.keys())[0] == "halos")
        ptype = 'halos'
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files,key=attrgetter("filename")):
            pcount = data_file.header['num_halos']
            if pcount == 0:
                continue
            with open(data_file.filename, "rb") as f:
                pos = data_file._get_particle_positions(ptype, f=f)
                yield "halos", (pos[:, i] for i in range(3))

    def _read_particle_fields(self, chunks, ptf, selector):
        # Now we have all the sizes, and we can allocate
        chunks = list(chunks)
        data_files = set([])
        # Only support halo reading for now.
        assert(len(ptf) == 1)
        assert(list(ptf.keys())[0] == "halos")
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files,key=attrgetter("filename")):
            pcount = data_file.header['num_halos']
            if pcount == 0:
                continue
            with open(data_file.filename, "rb") as f:
                for ptype, field_list in sorted(ptf.items()):
                    pos = data_file._get_particle_positions(ptype, f=f)
                    x, y, z = (pos[:, i] for i in range(3))
                    mask = selector.select_points(x, y, z, 0.0)
                    del x, y, z
                    f.seek(data_file._position_offset, os.SEEK_SET)
                    halos = np.fromfile(f, dtype=self._halo_dt, count = pcount)
                    if mask is None: continue
                    for field in field_list:
                        data = halos[field][mask].astype("float64")
                        yield (ptype, field), data

    def _initialize_index(self, data_file, regions):
        pcount = data_file.header["num_halos"]
        morton = np.empty(pcount, dtype='uint64')
        mylog.debug("Initializing index % 5i (% 7i particles)",
                    data_file.file_id, pcount)
        if pcount == 0:
            return morton
        ind = 0
        ptype = 'halos'
        with open(data_file.filename, "rb") as f:
            pos = data_file._get_particle_positions(ptype, f=f)
            pos = data_file.ds.arr(pos, "code_length")
            if np.any(pos.min(axis=0) < self.ds.domain_left_edge) or \
               np.any(pos.max(axis=0) > self.ds.domain_right_edge):
                raise YTDomainOverflow(pos.min(axis=0),
                                       pos.max(axis=0),
                                       self.ds.domain_left_edge,
                                       self.ds.domain_right_edge)
            regions.add_data_file(pos, data_file.file_id)
            morton[ind:ind+pos.shape[0]] = compute_morton(
                pos[:,0], pos[:,1], pos[:,2],
                data_file.ds.domain_left_edge,
                data_file.ds.domain_right_edge)
        return morton

    def _count_particles(self, data_file):
        return {'halos': data_file.header['num_halos']}

    def _identify_fields(self, data_file):
        fields = [("halos", f) for f in self._halo_dt.fields if
                  "padding" not in f]
        return fields, {}



"""
RAMSES-specific hilbert ordering routines.

These functions were translated from their original files from the
RAMSES project with the agreement of the original developer. See
https://bitbucket.org/rteyssie/ramses.


"""
from __future__ import print_function, absolute_import

#-----------------------------------------------------------------------------
# Copyright (c) yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

def hilbert3d(X, bit_length):
    '''Compute the order using Hilbert indexing.

    Arguments
    ---------
    * X: (N, ndim) float array
      The positions
    * bit_length: integer
      The bit_length for the indexing.
    '''
    X = np.atleast_2d(X)
    state_diagram = np.array([
        1, 2, 3, 2, 4, 5, 3, 5,
        0, 1, 3, 2, 7, 6, 4, 5,
        2, 6, 0, 7, 8, 8, 0, 7,
        0, 7, 1, 6, 3, 4, 2, 5,
        0, 9,10, 9, 1, 1,11,11,
        0, 3, 7, 4, 1, 2, 6, 5,
        6, 0, 6,11, 9, 0, 9, 8,
        2, 3, 1, 0, 5, 4, 6, 7,
        11,11, 0, 7, 5, 9, 0, 7,
        4, 3, 5, 2, 7, 0, 6, 1,
        4, 4, 8, 8, 0, 6,10, 6,
        6, 5, 1, 2, 7, 4, 0, 3,
        5, 7, 5, 3, 1, 1,11,11,
        4, 7, 3, 0, 5, 6, 2, 1,
        6, 1, 6,10, 9, 4, 9,10,
        6, 7, 5, 4, 1, 0, 2, 3,
        10, 3, 1, 1,10, 3, 5, 9,
        2, 5, 3, 4, 1, 6, 0, 7,
        4, 4, 8, 8, 2, 7, 2, 3,
        2, 1, 5, 6, 3, 0, 4, 7,
        7, 2,11, 2, 7, 5, 8, 5,
        4, 5, 7, 6, 3, 2, 0, 1,
        10, 3, 2, 6,10, 3, 4, 4,
        6, 1, 7, 0, 5, 2, 4, 3]).reshape(12, 2, 8).T

    x_bit_mask, y_bit_mask, z_bit_mask = [np.zeros(bit_length, dtype=bool) for _ in range(3)]
    i_bit_mask = np.zeros(3*bit_length, dtype=bool)

    npoint = X.shape[0]
    order = np.zeros(npoint)

    # Convert positions to binary
    for ip in range(npoint):
        for i in range(bit_length):
            mask = 0b01 << i
            x_bit_mask[i] = X[ip, 0] & mask
            y_bit_mask[i] = X[ip, 1] & mask
            z_bit_mask[i] = X[ip, 2] & mask

        for i in range(bit_length):
            # Interleave bits
            i_bit_mask[3*i+2] = x_bit_mask[i]
            i_bit_mask[3*i+1] = y_bit_mask[i]
            i_bit_mask[3*i  ] = z_bit_mask[i]

        # Build Hilbert ordering using state diagram
        cstate = 0
        for i in range(bit_length-1, -1, -1):
            sdigit = (4 * i_bit_mask[3*i+2] +
                      2 * i_bit_mask[3*i+1] +
                      1 * i_bit_mask[3*i  ])
            nstate = state_diagram[sdigit, 0, cstate]
            hdigit = state_diagram[sdigit, 1, cstate]

            i_bit_mask[3*i+2] = hdigit & 0b100
            i_bit_mask[3*i+1] = hdigit & 0b010
            i_bit_mask[3*i  ] = hdigit & 0b001

            cstate = nstate

        # Compute ordering
        for i in range(3*bit_length):
            order[ip] = order[ip] + i_bit_mask[i]*2**i

    return order

def get_cpu_list(ds, X):
    '''
    Return the list of the CPU intersecting with the positions
    given. Note that it will be 0-indexed.

    Parameters
    ----------
    * ds: Dataset
      The dataset containing the information
    * X: (N, ndim) float array
      An array containing positions. They should be between 0 and 1.
    '''
    X = np.atleast_2d(X)
    if X.shape[1] != 3:
        raise NotImplementedError('This function is only implemented in 3D.')

    levelmax = ds.parameters['levelmax']
    ncpu = ds.parameters['ncpu']
    ndim = ds.parameters['ndim']

    xmin, ymin, zmin = X.min(axis=0)
    xmax, ymax, zmax = X.max(axis=0)

    dmax = max(xmax-xmin, ymax-ymin, zmax-zmin)
    ilevel = 0
    deltax = dmax * 2

    while deltax >= dmax:
        ilevel += 1
        deltax = 0.5**ilevel

    lmin = ilevel
    bit_length = lmin - 1
    maxdom = 2**bit_length

    imin, imax, jmin, jmax, kmin, kmax = 0, 0, 0, 0, 0, 0
    if bit_length > 0:
        imin = int(xmin * maxdom)
        imax = imin + 1
        jmin = int(ymin * maxdom)
        jmax = jmin + 1
        kmin = int(zmin * maxdom)
        kmax = kmin + 1


    dkey = (2**(levelmax+1) / maxdom)**ndim
    ndom = 1
    if (bit_length > 0): ndom = 8

    idom, jdom, kdom = [np.zeros(8, dtype=int) for _ in range(3)]

    idom[0], idom[1] = imin, imax
    idom[2], idom[3] = imin, imax
    idom[4], idom[5] = imin, imax
    idom[6], idom[7] = imin, imax

    jdom[0], jdom[1] = jmin, jmin
    jdom[2], jdom[3] = jmax, jmax
    jdom[4], jdom[5] = jmin, jmin
    jdom[6], jdom[7] = jmax, jmax

    kdom[0], kdom[1] = kmin, kmin
    kdom[2], kdom[3] = kmin, kmin
    kdom[4], kdom[5] = kmax, kmax
    kdom[6], kdom[7] = kmax, kmax

    bounding_min, bounding_max = np.zeros(ndom), np.zeros(ndom)
    for i in range(ndom):
        if bit_length > 0:
            order_min = hilbert3d([idom[i], jdom[i], kdom[i]], bit_length)
        else:
            order_min = 0
        bounding_min[i] = (order_min  )*dkey
        bounding_max[i] = (order_min+1)*dkey

    bound_key = {}
    for icpu in range(1, ncpu+1):
        bound_key[icpu-1], bound_key[icpu] = ds.hilbert_indices[icpu]

    cpu_min, cpu_max = [np.zeros(ncpu + 1, dtype=np.int) for _ in range(2)]
    for icpu in range(1, ncpu+1):
        for i in range(ndom):
            if (bound_key[icpu - 1] <= bounding_min[i] and
                bound_key[icpu    ] >  bounding_min[i]):
                cpu_min[i] = icpu-1
            if (bound_key[icpu - 1] <  bounding_max[i] and
                bound_key[icpu    ] >= bounding_max[i]):
                cpu_max[i] = icpu

    ncpu_read = 0
    cpu_list = []
    cpu_read = np.zeros(ncpu, dtype=np.bool)
    for i in range(ndom):
        for j in range(cpu_min[i], cpu_max[i]):
            if not cpu_read[j]:
                ncpu_read += 1
                cpu_list.append(j)
                cpu_read[j] = True

    return sorted(cpu_list)



"""
Athena++-specific fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer
from yt.utilities.physical_constants import \
    kboltz, mh

b_units = "code_magnetic"
pres_units = "code_mass/(code_length*code_time**2)"
rho_units = "code_mass / code_length**3"
vel_units = "code_length / code_time"

def velocity_field(j):
    def _velocity(field, data):
        return data["athena_pp", "mom%d" % j]/data["athena_pp","dens"]
    return _velocity

class AthenaPPFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("rho", (rho_units, ["density"], None)),
        ("dens", (rho_units, ["density"], None)),
        ("Bcc1", (b_units, [], None)),
        ("Bcc2", (b_units, [], None)),
        ("Bcc3", (b_units, [], None)),
    )

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import \
            setup_magnetic_field_aliases
        unit_system = self.ds.unit_system
        # Add velocity fields
        vel_prefix = "velocity"
        for i, comp in enumerate(self.ds.coordinates.axis_order):
            vel_field = ("athena_pp", "vel%d" % (i+1))
            mom_field = ("athena_pp", "mom%d" % (i+1))
            if vel_field in self.field_list:
                self.add_output_field(vel_field, sampling_type="cell", units="code_length/code_time")
                self.alias(("gas","%s_%s" % (vel_prefix, comp)), vel_field,
                           units=unit_system["velocity"])
            elif mom_field in self.field_list:
                self.add_output_field(mom_field, sampling_type="cell",
                                      units="code_mass/code_time/code_length**2")
                self.add_field(("gas","%s_%s" % (vel_prefix, comp)), sampling_type="cell",
                               function=velocity_field(i+1), units=unit_system["velocity"])
        # Figure out thermal energy field
        if ("athena_pp","press") in self.field_list:
            self.add_output_field(("athena_pp","press"), sampling_type="cell",
                                  units=pres_units)
            self.alias(("gas","pressure"),("athena_pp","press"),
                       units=unit_system["pressure"])
            def _thermal_energy(field, data):
                return data["athena_pp","press"] / \
                       (data.ds.gamma-1.)/data["athena_pp","rho"]
            self.add_field(("gas","thermal_energy"), sampling_type="cell",
                           function=_thermal_energy,
                           units=unit_system["specific_energy"])
        elif ("athena_pp","Etot") in self.field_list:
            self.add_output_field(("athena_pp","Etot"), sampling_type="cell",
                                  units=pres_units)
            def _thermal_energy(field, data):
                eint = data["athena_pp", "Etot"] - data["gas","kinetic_energy"]
                if ("athena_pp", "B1") in self.field_list:
                    eint -= data["gas","magnetic_energy"]
                return eint/data["athena_pp","dens"]
            self.add_field(("gas","thermal_energy"), sampling_type="cell",
                           function=_thermal_energy,
                           units=unit_system["specific_energy"])
        # Add temperature field
        def _temperature(field, data):
            if data.has_field_parameter("mu"):
                mu = data.get_field_parameter("mu")
            else:
                mu = 0.6
            return (data["gas","pressure"]/data["gas","density"])*mu*mh/kboltz
        self.add_field(("gas", "temperature"), sampling_type="cell", function=_temperature,
                       units=unit_system["temperature"])

        setup_magnetic_field_aliases(self, "athena_pp", ["Bcc%d" % ax for ax in (1,2,3)])



import numpy as np
from yt.testing import \
    assert_almost_equal, \
    assert_equal
from yt.utilities.particle_generator import \
    WithDensityParticleGenerator, \
    LatticeParticleGenerator, \
    FromListParticleGenerator
from yt.frontends.stream.api import load_uniform_grid, refine_amr
import yt.utilities.initial_conditions as ic
import yt.utilities.flagging_methods as fm
from yt.units.yt_array import uconcatenate

def test_particle_generator():
    # First generate our dataset
    domain_dims = (32, 32, 32)
    dens = np.zeros(domain_dims) + 0.1
    temp = 4.*np.ones(domain_dims)
    fields = {"density": (dens, 'code_mass/code_length**3'),
              "temperature": (temp, 'K')}
    ug = load_uniform_grid(fields, domain_dims, 1.0)
    fo = [ic.BetaModelSphere(1.0,0.1,0.5,[0.5,0.5,0.5],{"density":(10.0)})]
    rc = [fm.flagging_method_registry["overdensity"](4.0)]
    ds = refine_amr(ug, rc, fo, 3)

    # Now generate particles from density

    field_list = [("io", "particle_position_x"),
                  ("io", "particle_position_y"),
                  ("io", "particle_position_z"),
                  ("io", "particle_index"),
                  ("io", "particle_gas_density")]
    num_particles = 10000
    field_dict = {("gas", "density"): ("io", "particle_gas_density")}
    sphere = ds.sphere(ds.domain_center, 0.45)

    particles1 = WithDensityParticleGenerator(ds, sphere, num_particles, field_list)
    particles1.assign_indices()
    particles1.map_grid_fields_to_particles(field_dict)

    # Test to make sure we ended up with the right number of particles per grid
    particles1.apply_to_stream()
    particles_per_grid1 = [grid.NumberOfParticles for grid in ds.index.grids]
    assert_equal(particles_per_grid1, particles1.NumberOfParticles)
    particles_per_grid1 = [len(grid["particle_position_x"]) for grid in ds.index.grids]
    assert_equal(particles_per_grid1, particles1.NumberOfParticles)

    tags = uconcatenate([grid["particle_index"] for grid in ds.index.grids])
    assert(np.unique(tags).size == num_particles)

    del tags

    # Set up a lattice of particles
    pdims = np.array([32,32,32])
    def new_indices():
        # We just add new indices onto the existing ones
        return np.arange((np.product(pdims)))+num_particles
    le = np.array([0.25,0.25,0.25])
    re = np.array([0.75,0.75,0.75])

    particles2 = LatticeParticleGenerator(ds, pdims, le, re, field_list)
    particles2.assign_indices(function=new_indices)
    particles2.map_grid_fields_to_particles(field_dict)

    # Test lattice positions
    xpos = np.unique(particles2["io", "particle_position_x"])
    ypos = np.unique(particles2["io", "particle_position_y"])
    zpos = np.unique(particles2["io", "particle_position_z"])

    xpred = np.linspace(le[0],re[0],num=pdims[0],endpoint=True)
    ypred = np.linspace(le[1],re[1],num=pdims[1],endpoint=True)
    zpred = np.linspace(le[2],re[2],num=pdims[2],endpoint=True)

    assert_almost_equal(xpos, xpred)
    assert_almost_equal(ypos, ypred)
    assert_almost_equal(zpos, zpred)

    del xpos, ypos, zpos
    del xpred, ypred, zpred

    # Test the number of particles again
    particles2.apply_to_stream()
    particles_per_grid2 = [grid.NumberOfParticles for grid in ds.index.grids]
    assert_equal(particles_per_grid2, particles1.NumberOfParticles+particles2.NumberOfParticles)

    [grid.field_data.clear() for grid in ds.index.grids]
    particles_per_grid2 = [len(grid["particle_position_x"]) for grid in ds.index.grids]
    assert_equal(particles_per_grid2, particles1.NumberOfParticles+particles2.NumberOfParticles)

    # Test the uniqueness of tags
    tags = np.concatenate([grid["particle_index"] for grid in ds.index.grids])
    tags.sort()
    assert_equal(tags, np.arange((np.product(pdims)+num_particles)))

    del tags

    # Now dump all of these particle fields out into a dict
    pdata = {}
    dd = ds.all_data()
    for field in field_list:
        pdata[field] = dd[field]

    # Test the "from-list" generator and particle field overwrite
    num_particles3 = num_particles+np.product(pdims)
    particles3 = FromListParticleGenerator(ds, num_particles3, pdata)
    particles3.apply_to_stream(overwrite=True)

    # Test the number of particles again
    particles_per_grid3 = [grid.NumberOfParticles for grid in ds.index.grids]
    assert_equal(particles_per_grid3, particles1.NumberOfParticles+particles2.NumberOfParticles)
    particles_per_grid2 = [len(grid["particle_position_z"]) for grid in ds.index.grids]
    assert_equal(particles_per_grid3, particles1.NumberOfParticles+particles2.NumberOfParticles)
    assert_equal(particles_per_grid2, particles_per_grid3)

    # Test adding in particles with a different particle type

    num_star_particles = 20000
    pdata2 = {("star", "particle_position_x"): np.random.uniform(size=num_star_particles),
              ("star", "particle_position_y"): np.random.uniform(size=num_star_particles),
              ("star", "particle_position_z"): np.random.uniform(size=num_star_particles)}

    particles4 = FromListParticleGenerator(ds, num_star_particles, pdata2, ptype="star")
    particles4.apply_to_stream()

    dd = ds.all_data()
    assert dd["star", "particle_position_x"].size == num_star_particles
    assert dd["io", "particle_position_x"].size == num_particles3
    assert dd["all", "particle_position_x"].size == num_star_particles+num_particles3


    del pdata
    del pdata2
    del ds
    del particles1
    del particles2
    del particles4
    del fields
    del dens
    del temp




"""
Integrator classes to deal with interpolation and integration of input spectral
bins.  Currently only supports Cloudy and APEC-style data.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.utilities.on_demand_imports import _h5py as h5py
import numpy as np
import os

from yt.config import ytcfg
from yt.fields.derived_field import DerivedField
from yt.funcs import \
    mylog, \
    only_on_root, \
    issue_deprecation_warning, \
    parse_h5_attr
from yt.utilities.exceptions import YTFieldNotFound
from yt.utilities.exceptions import YTException
from yt.utilities.linear_interpolators import \
    UnilinearFieldInterpolator, BilinearFieldInterpolator
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.cosmology import Cosmology

data_version = {"cloudy": 2,
                "apec": 2}

data_url = "http://yt-project.org/data"

def _get_data_file(table_type, data_dir=None):
    data_file = "%s_emissivity_v%d.h5" % (table_type, data_version[table_type])
    if data_dir is None:
        supp_data_dir = ytcfg.get("yt", "supp_data_dir")
        data_dir = supp_data_dir if os.path.exists(supp_data_dir) else "."
    data_path = os.path.join(data_dir, data_file)
    if not os.path.exists(data_path):
        msg = "Failed to find emissivity data file %s! " % data_file + \
            "Please download from http://yt-project.org/data!"
        mylog.error(msg)
        raise IOError(msg)
    return data_path

class EnergyBoundsException(YTException):
    def __init__(self, lower, upper):
        self.lower = lower
        self.upper = upper

    def __str__(self):
        return "Energy bounds are %e to %e keV." % \
          (self.lower, self.upper)

class ObsoleteDataException(YTException):
    def __init__(self, table_type):
        data_file = "%s_emissivity_v%d.h5" % (table_type, data_version[table_type])
        self.msg = "X-ray emissivity data is out of date.\n"
        self.msg += "Download the latest data from %s/%s." % (data_url, data_file)

    def __str__(self):
        return self.msg

class XrayEmissivityIntegrator(object):
    r"""Class for making X-ray emissivity fields. Uses hdf5 data tables
    generated from Cloudy and AtomDB/APEC.

    Initialize an XrayEmissivityIntegrator object.

    Parameters
    ----------
    table_type : string
        The type of data to use when computing the emissivity values. If "cloudy",
        a file called "cloudy_emissivity.h5" is used, for photoionized
        plasmas. If, "apec", a file called "apec_emissivity.h5" is used for 
        collisionally ionized plasmas. These files contain emissivity tables 
        for primordial elements and for metals at solar metallicity for the 
        energy range 0.1 to 100 keV.
    redshift : float, optional
        The cosmological redshift of the source of the field. Default: 0.0.
    data_dir : string, optional
        The location to look for the data table in. If not supplied, the file
        will be looked for in the location of the YT_DEST environment variable
        or in the current working directory.
    use_metals : boolean, optional
        If set to True, the emissivity will include contributions from metals.
        Default: True
    """
    def __init__(self, table_type, redshift=0.0, data_dir=None, use_metals=True):

        filename = _get_data_file(table_type, data_dir=data_dir)
        only_on_root(mylog.info, "Loading emissivity data from %s." % filename)
        in_file = h5py.File(filename, "r")
        if "info" in in_file.attrs:
            only_on_root(mylog.info, parse_h5_attr(in_file, "info"))
        if parse_h5_attr(in_file, "version") != data_version[table_type]:
            raise ObsoleteDataException(table_type)
        else:
            only_on_root(mylog.info, "X-ray '%s' emissivity data version: %s." % \
                         (table_type, parse_h5_attr(in_file, "version")))

        self.log_T = in_file["log_T"][:]
        self.emissivity_primordial = in_file["emissivity_primordial"][:]
        if "log_nH" in in_file:
            self.log_nH = in_file["log_nH"][:]
        if use_metals:
            self.emissivity_metals = in_file["emissivity_metals"][:]
        self.ebin = YTArray(in_file["E"], "keV")
        in_file.close()
        self.dE = np.diff(self.ebin)
        self.emid = 0.5*(self.ebin[1:]+self.ebin[:-1]).to("erg")
        self.redshift = redshift

    def get_interpolator(self, data_type, e_min, e_max, energy=True):
        data = getattr(self, "emissivity_%s" % data_type)
        if not energy:
            data = data[..., :] / self.emid.v
        e_min = YTQuantity(e_min, "keV")*(1.0+self.redshift)
        e_max = YTQuantity(e_max, "keV")*(1.0+self.redshift)
        if (e_min - self.ebin[0]) / e_min < -1e-3 or \
          (e_max - self.ebin[-1]) / e_max > 1e-3:
            raise EnergyBoundsException(self.ebin[0], self.ebin[-1])
        e_is, e_ie = np.digitize([e_min, e_max], self.ebin)
        e_is = np.clip(e_is - 1, 0, self.ebin.size - 1)
        e_ie = np.clip(e_ie, 0, self.ebin.size - 1)

        my_dE = self.dE[e_is: e_ie].copy()
        # clip edge bins if the requested range is smaller
        my_dE[0] -= e_min - self.ebin[e_is]
        my_dE[-1] -= self.ebin[e_ie] - e_max

        interp_data = (data[..., e_is:e_ie]*my_dE).sum(axis=-1)
        if data.ndim == 2:
            emiss = UnilinearFieldInterpolator(np.log10(interp_data),
                                               [self.log_T[0],  self.log_T[-1]],
                                               "log_T", truncate=True)
        else:
            emiss = BilinearFieldInterpolator(np.log10(interp_data),
                                              [self.log_nH[0], self.log_nH[-1],
                                               self.log_T[0],  self.log_T[-1]],
                                              ["log_nH", "log_T"], truncate=True)

        return emiss

def add_xray_emissivity_field(ds, e_min, e_max, redshift=0.0,
                              metallicity=("gas", "metallicity"), 
                              table_type="cloudy", data_dir=None,
                              cosmology=None, **kwargs):
    r"""Create X-ray emissivity fields for a given energy range.

    Parameters
    ----------
    e_min : float
        The minimum energy in keV for the energy band.
    e_min : float
        The maximum energy in keV for the energy band.
    redshift : float, optional
        The cosmological redshift of the source of the field. Default: 0.0.
    metallicity : str or tuple of str or float, optional
        Either the name of a metallicity field or a single floating-point
        number specifying a spatially constant metallicity. Must be in
        solar units. If set to None, no metals will be assumed. Default: 
        ("gas", "metallicity")
    table_type : string, optional
        The type of emissivity table to be used when creating the fields. 
        Options are "cloudy" or "apec". Default: "cloudy"
    data_dir : string, optional
        The location to look for the data table in. If not supplied, the file
        will be looked for in the location of the YT_DEST environment variable
        or in the current working directory.
    cosmology : :class:`~yt.utilities.cosmology.Cosmology`, optional
        If set and redshift > 0.0, this cosmology will be used when computing the
        cosmological dependence of the emission fields. If not set, yt's default
        LCDM cosmology will be used.

    This will create three fields:

    "xray_emissivity_{e_min}_{e_max}_keV" (erg s^-1 cm^-3)
    "xray_luminosity_{e_min}_{e_max}_keV" (erg s^-1)
    "xray_photon_emissivity_{e_min}_{e_max}_keV" (photons s^-1 cm^-3)

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("sloshing_nomag2_hdf5_plt_cnt_0100")
    >>> yt.add_xray_emissivity_field(ds, 0.5, 2)
    >>> p = yt.ProjectionPlot(ds, 'x', "xray_emissivity_0.5_2_keV")
    >>> p.save()
    """
    # The next several if constructs are for backwards-compatibility
    if "constant_metallicity" in kwargs:
        issue_deprecation_warning("The \"constant_metallicity\" parameter is deprecated. Set "
                                  "the \"metallicity\" parameter to a constant float value instead.")
        metallicity = kwargs["constant_metallicity"]
    if "with_metals" in kwargs:
        issue_deprecation_warning("The \"with_metals\" parameter is deprecated. Use the "
                                  "\"metallicity\" parameter to choose a constant or "
                                  "spatially varying metallicity.")
        if kwargs["with_metals"] and isinstance(metallicity, float):
            raise RuntimeError("\"with_metals=True\", but you specified a constant metallicity!")
        if not kwargs["with_metals"] and not isinstance(metallicity, float):
            raise RuntimeError("\"with_metals=False\", but you didn't specify a constant metallicity!")
    if not isinstance(metallicity, float) and metallicity is not None:
        try:
            metallicity = ds._get_field_info(*metallicity)
        except YTFieldNotFound:
            raise RuntimeError("Your dataset does not have a {} field! ".format(metallicity) +
                               "Perhaps you should specify a constant metallicity instead?")

    my_si = XrayEmissivityIntegrator(table_type, data_dir=data_dir, redshift=redshift)

    em_0 = my_si.get_interpolator("primordial", e_min, e_max)
    emp_0 = my_si.get_interpolator("primordial", e_min, e_max, energy=False)
    if metallicity is not None:
        em_Z = my_si.get_interpolator("metals", e_min, e_max)
        emp_Z = my_si.get_interpolator("metals", e_min, e_max, energy=False)

    def _emissivity_field(field, data):
        with np.errstate(all='ignore'):
            dd = {"log_nH": np.log10(data["gas", "H_nuclei_density"]),
                  "log_T": np.log10(data["gas", "temperature"])}

        my_emissivity = np.power(10, em_0(dd))
        if metallicity is not None:
            if isinstance(metallicity, DerivedField):
                my_Z = data[metallicity.name]
            else:
                my_Z = metallicity
            my_emissivity += my_Z * np.power(10, em_Z(dd))

        my_emissivity[np.isnan(my_emissivity)] = 0

        return data["gas","H_nuclei_density"]**2 * \
            YTArray(my_emissivity, "erg*cm**3/s")

    emiss_name = "xray_emissivity_%s_%s_keV" % (e_min, e_max)
    ds.add_field(("gas", emiss_name), function=_emissivity_field,
                 display_name=r"\epsilon_{X} (%s-%s keV)" % (e_min, e_max),
                 sampling_type="cell", units="erg/cm**3/s")

    def _luminosity_field(field, data):
        return data[emiss_name] * data["cell_volume"]

    lum_name = "xray_luminosity_%s_%s_keV" % (e_min, e_max)
    ds.add_field(("gas", lum_name), function=_luminosity_field,
                 display_name=r"\rm{L}_{X} (%s-%s keV)" % (e_min, e_max),
                 sampling_type="cell", units="erg/s")

    def _photon_emissivity_field(field, data):
        dd = {"log_nH": np.log10(data["gas", "H_nuclei_density"]),
              "log_T": np.log10(data["gas", "temperature"])}

        my_emissivity = np.power(10, emp_0(dd))
        if metallicity is not None:
            if isinstance(metallicity, DerivedField):
                my_Z = data[metallicity.name]
            else:
                my_Z = metallicity
            my_emissivity += my_Z * np.power(10, emp_Z(dd))

        return data["gas", "H_nuclei_density"]**2 * \
            YTArray(my_emissivity, "photons*cm**3/s")

    phot_name = "xray_photon_emissivity_%s_%s_keV" % (e_min, e_max)
    ds.add_field(("gas", phot_name), function=_photon_emissivity_field,
                 display_name=r"\epsilon_{X} (%s-%s keV)" % (e_min, e_max),
                 sampling_type="cell", units="photons/cm**3/s")

    fields = [emiss_name, lum_name, phot_name]

    if redshift > 0.0:

        if cosmology is None:
            if hasattr(ds, "cosmology"):
                cosmology = ds.cosmology
            else:
                cosmology = Cosmology()

        D_L = cosmology.luminosity_distance(0.0, redshift)
        angular_scale = 1.0/cosmology.angular_scale(0.0, redshift)
        dist_fac = 1.0/(4.0*np.pi*D_L*D_L*angular_scale*angular_scale)

        ei_name = "xray_intensity_%s_%s_keV" % (e_min, e_max)
        def _intensity_field(field, data):
            I = dist_fac*data[emiss_name]
            return I.in_units("erg/cm**3/s/arcsec**2")
        ds.add_field(("gas", ei_name), function=_intensity_field,
                     display_name=r"I_{X} (%s-%s keV)" % (e_min, e_max),
                     sampling_type="cell", units="erg/cm**3/s/arcsec**2")

        i_name = "xray_photon_intensity_%s_%s_keV" % (e_min, e_max)
        def _photon_intensity_field(field, data):
            I = (1.0+redshift)*dist_fac*data[phot_name]
            return I.in_units("photons/cm**3/s/arcsec**2")
        ds.add_field(("gas", i_name), function=_photon_intensity_field,
                     display_name=r"I_{X} (%s-%s keV)" % (e_min, e_max),
                     sampling_type="cell", units="photons/cm**3/s/arcsec**2")

        fields += [ei_name, i_name]

    [mylog.info("Adding %s field." % field) for field in fields]

    return fields



"""
Testsuite for ProfilePlot and PhasePlot



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------
import os
import tempfile
import shutil
import unittest
import yt
from nose.plugins.attrib import attr
from yt.data_objects.profiles import create_profile
from yt.testing import \
    fake_random_ds, \
    assert_array_almost_equal, \
    assert_allclose_units, ANSWER_TEST_TAG
from yt.visualization.profile_plotter import \
    ProfilePlot, PhasePlot
from yt.utilities.answer_testing.framework import \
    PhasePlotAttributeTest, GenericImageTest

ATTR_ARGS = {"annotate_text": [(((5e-29, 5e7), "Hello YT"), {}),
                               (((5e-29, 5e7), "Hello YT"), {'color': 'b'})],

             "set_title": [(('cell_mass', 'A phase plot.'), {})],
             "set_log": [(('cell_mass', False), {})],
             "set_unit": [(('cell_mass', 'Msun'), {})],
             "set_xlim": [((1e-27, 1e-24), {})],
             "set_ylim": [((1e2, 1e6), {})]}


def compare(ds, plot, test_prefix, test_name, decimals=12):
    def image_from_plot(filename_prefix):
        return plot.save(filename_prefix)

    image_from_plot.__name__ = "profile_{}".format(test_prefix)
    test = GenericImageTest(ds, image_from_plot, decimals)
    test.prefix = test_prefix
    test.answer_name = test_name
    return test


@attr(ANSWER_TEST_TAG)
def test_phase_plot_attributes():
    '''

    This iterates over the all the plot modification functions in
    ATTR_ARGS. Each time, it compares the images produced by
    PhasePlot to the gold standard.


    '''

    x_field = 'density'
    y_field = 'temperature'
    z_field = 'cell_mass'
    decimals = 12
    ds = fake_random_ds(16, fields=('density', 'temperature'))
    for attr_name in ATTR_ARGS.keys():
        for args in ATTR_ARGS[attr_name]:
            test = PhasePlotAttributeTest(ds, x_field, y_field, z_field,
                                          attr_name, args, decimals)
            test.prefix = "%s_%s" % (attr_name, args)
            test.answer_name = "phase_plot_attributes"
            yield test


@attr(ANSWER_TEST_TAG)
def test_profile_plot():
    fields = ('density', 'temperature', 'velocity_x', 'velocity_y',
              'velocity_z')
    units = ('g/cm**3', 'K', 'cm/s', 'cm/s', 'cm/s')
    test_ds = fake_random_ds(16, fields=fields, units=units)
    regions = [test_ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3), test_ds.all_data()]
    pr_fields = [('density', 'temperature'), ('density', 'velocity_x'),
                 ('temperature', 'cell_mass'), ('density', 'radius'),
                 ('velocity_magnitude', 'cell_mass')]
    profiles = []
    for reg in regions:
        for x_field, y_field in pr_fields:
            profiles.append(ProfilePlot(reg, x_field, y_field))
            profiles.append(ProfilePlot(reg, x_field, y_field,
                                        fractional=True, accumulation=True))
            p1d = create_profile(reg, x_field, y_field)
            profiles.append(ProfilePlot.from_profiles(p1d))
    p1 = create_profile(test_ds.all_data(), 'density', 'temperature')
    p2 = create_profile(test_ds.all_data(), 'density', 'velocity_x')
    profiles.append(ProfilePlot.from_profiles(
        [p1, p2], labels=['temperature', 'velocity']))
    profiles[0]._repr_html_()
    for idx, plot in enumerate(profiles):
        test_prefix = "%s_%s" % (plot.plots.keys(), idx)
        yield compare(test_ds, plot, test_prefix=test_prefix,
                      test_name="profile_plots")


@attr(ANSWER_TEST_TAG)
def test_phase_plot():
    fields = ('density', 'temperature', 'velocity_x', 'velocity_y',
              'velocity_z')
    units = ('g/cm**3', 'K', 'cm/s', 'cm/s', 'cm/s')
    test_ds = fake_random_ds(16, fields=fields, units=units)
    regions = [test_ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3), test_ds.all_data()]
    phases = []
    ph_fields = [('density', 'temperature', 'cell_mass'),
                 ('density', 'velocity_x', 'cell_mass'),
                 ('radius', 'temperature', 'velocity_magnitude')]
    for reg in regions:
        for x_field, y_field, z_field in ph_fields:
            # set n_bins to [16, 16] since matplotlib's postscript
            # renderer is slow when it has to write a lot of polygons
            phases.append(PhasePlot(reg, x_field, y_field, z_field,
                                    x_bins=16, y_bins=16))
            phases.append(PhasePlot(reg, x_field, y_field, z_field,
                                    fractional=True, accumulation=True,
                                    x_bins=16, y_bins=16))
            p2d = create_profile(reg, [x_field, y_field], z_field,
                                 n_bins=[16, 16])
            phases.append(PhasePlot.from_profile(p2d))
    pp = PhasePlot(test_ds.all_data(), 'density', 'temperature', 'cell_mass')
    pp.set_xlim(0.3, 0.8)
    pp.set_ylim(0.4, 0.6)
    pp._setup_plots()
    xlim = pp.plots['cell_mass'].axes.get_xlim()
    ylim = pp.plots['cell_mass'].axes.get_ylim()
    assert_array_almost_equal(xlim, (0.3, 0.8))
    assert_array_almost_equal(ylim, (0.4, 0.6))
    phases.append(pp)
    phases[0]._repr_html_()
    for idx, plot in enumerate(phases):
        test_prefix = "%s_%s" % (plot.plots.keys(), idx)
        yield compare(test_ds, plot, test_prefix=test_prefix,
                      test_name="phase_plots")


@attr(ANSWER_TEST_TAG)
def test_profile_plot_multiple_field_multiple_plot():
    fields = ('density', 'temperature', 'dark_matter_density')
    ds = fake_random_ds(16, fields=fields)
    sphere = ds.sphere("max", (1.0, "Mpc"))
    profiles = []
    profiles.append(yt.create_profile(sphere, ["radius"],
                                      fields=["density"], n_bins=32))
    profiles.append(yt.create_profile(sphere, ["radius"],
                                      fields=["density"], n_bins=64))
    profiles.append(yt.create_profile(sphere, ["radius"],
                                      fields=["dark_matter_density"],
                                      n_bins=64))

    plot = yt.ProfilePlot.from_profiles(profiles)
    yield compare(ds, plot, test_prefix=plot.plots.keys(),
                  test_name="profile_plot_multiple_field_multiple_plot")


def test_set_units():
    fields = ('density', 'temperature')
    units = ('g/cm**3', 'K',)
    ds = fake_random_ds(16, fields=fields, units=units)
    sp = ds.sphere("max", (1.0, "Mpc"))
    p1 = yt.ProfilePlot(sp, "radius", ("gas", "density"))
    p2 = yt.PhasePlot(sp, ("gas", "density"), ("gas", "temperature"), "cell_mass")
    # make sure we can set the units using the tuple without erroring out
    p1.set_unit(("gas", "density"), "Msun/kpc**3")
    p2.set_unit(("gas", "temperature"), "R")


def test_set_labels():
    ds = fake_random_ds(16)
    ad = ds.all_data()
    plot = yt.ProfilePlot(ad, "radius", ["velocity_x", "density"], weight_field=None)
    # make sure we can set the labels without erroring out
    plot.set_ylabel("all", "test ylabel")
    plot.set_xlabel("test xlabel")


def test_create_from_dataset():
    ds = fake_random_ds(16)
    plot1 = yt.ProfilePlot(ds, "radius", ["velocity_x", "density"],
                           weight_field=None)
    plot2 = yt.ProfilePlot(ds.all_data(), "radius", ["velocity_x", "density"],
                           weight_field=None)
    assert_allclose_units(
        plot1.profiles[0]['density'], plot2.profiles[0]['density'])
    assert_allclose_units(
        plot1.profiles[0]['velocity_x'], plot2.profiles[0]['velocity_x'])

    plot1 = yt.PhasePlot(ds, 'density', 'velocity_x', 'cell_mass')
    plot2 = yt.PhasePlot(ds.all_data(), 'density', 'velocity_x', 'cell_mass')
    assert_allclose_units(
        plot1.profile['cell_mass'], plot2.profile['cell_mass'])


class TestAnnotations(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.curdir = os.getcwd()
        os.chdir(cls.tmpdir)

        ds = fake_random_ds(16)
        ad = ds.all_data()
        cls.fields = ["velocity_x", "velocity_y", "velocity_z"]
        cls.plot = yt.ProfilePlot(ad, "radius", cls.fields, weight_field=None)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.curdir)
        shutil.rmtree(cls.tmpdir)

    def test_annotations(self):
        # make sure we can annotate without erroring out
        # annotate the plot with only velocity_x
        self.plot.annotate_title("velocity_x plot", self.fields[0])
        self.plot.annotate_text(1e-1, 1e1, "Annotated velocity_x")

        # annotate the plots with velocity_y and velocity_z with
        # the same annotations
        self.plot.annotate_title("Velocity Plots (Y or Z)", self.fields[1:])
        self.plot.annotate_text(1e-1, 1e1, "Annotated vel_y, vel_z", self.fields[1:])
        self.plot.save()

    def test_annotations_wrong_fields(self):
        from yt.utilities.exceptions import YTFieldNotFound
        with self.assertRaises(YTFieldNotFound):
            self.plot.annotate_title("velocity_x plot", "wrong_field_name")

        with self.assertRaises(YTFieldNotFound):
            self.plot.annotate_text(1e-1, 1e1, "Annotated text", "wrong_field_name")


def test_phaseplot_set_log():
    fields = ('density', 'temperature')
    units = ('g/cm**3', 'K',)
    ds = fake_random_ds(16, fields=fields, units=units)
    sp = ds.sphere("max", (1.0, "Mpc"))
    p1 = yt.ProfilePlot(sp, "radius", ("gas", "density"))
    p2 = yt.PhasePlot(sp, ("gas", "density"), ("gas", "temperature"), "cell_mass")
    # make sure we can set the log-scaling using the tuple without erroring out
    p1.set_log(("gas", "density"), False)
    p2.set_log(("gas", "temperature"), False)
    assert p1.y_log["gas", "density"] is False
    assert p2.y_log is False

    # make sure we can set the log-scaling using a string without erroring out
    p1.set_log("density", True)
    p2.set_log("temperature", True)
    assert p1.y_log["gas", "density"] is True
    assert p2.y_log is True

    # make sure we can set the log-scaling using a field object
    p1.set_log(ds.fields.gas.density, False)
    p2.set_log(ds.fields.gas.temperature, False)
    assert p1.y_log["gas", "density"] is False
    assert p2.y_log is False



def test_phaseplot_showhide_colorbar_axes():
    fields = ('density', 'temperature')
    units = ('g/cm**3', 'K',)
    ds = fake_random_ds(16, fields=fields, units=units)
    ad = ds.all_data()
    plot = yt.PhasePlot(ad, ("gas", "density"), ("gas", "temperature"), "cell_mass")

    # make sure we can hide colorbar
    plot.hide_colorbar()
    with tempfile.NamedTemporaryFile(suffix='png') as f1:
        plot.save(f1.name)

    # make sure we can show colorbar
    plot.show_colorbar()
    with tempfile.NamedTemporaryFile(suffix='png') as f2:
        plot.save(f2.name)

    # make sure we can hide axes
    plot.hide_axes()
    with tempfile.NamedTemporaryFile(suffix='png') as f3:
        plot.save(f3.name)

    # make sure we can show axes
    plot.show_axes()
    with tempfile.NamedTemporaryFile(suffix='png') as f4:
        plot.save(f4.name)







"""
Projection class for the Sunyaev-Zeldovich effect. Requires SZpack (at least
version 1.1.1) to be downloaded and installed:

http://www.jb.man.ac.uk/~jchluba/Science/SZpack/SZpack.html

For details on the computations involved please refer to the following references:

Chluba, Nagai, Sazonov, Nelson, MNRAS, 2012, arXiv:1205.5778
Chluba, Switzer, Nagai, Nelson, MNRAS, 2012, arXiv:1211.3206
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.config import \
    ytcfg
from yt.utilities.physical_constants import sigma_thompson, clight, hcgs, kboltz, mh, Tcmb
from yt.funcs import fix_axis, get_pbar
from yt.visualization.volume_rendering.off_axis_projection import \
    off_axis_projection
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    communication_system, parallel_root_only
from yt import units
from yt.utilities.on_demand_imports import _astropy

import numpy as np

I0 = (2*(kboltz*Tcmb)**3/((hcgs*clight)**2)/units.sr).in_units("MJy/steradian")

try:
    import SZpack
except ImportError:
    pass

vlist = "xyz"
def setup_sunyaev_zeldovich_fields(ds):
    def _t_squared(field, data):
        return data["gas","density"]*data["gas","kT"]*data["gas","kT"]
    ds.add_field(("gas", "t_squared"), function = _t_squared,
                 units="g*keV**2/cm**3")

    def _beta_par_squared(field, data):
        return data["gas","beta_par"]**2/data["gas","density"]
    ds.add_field(("gas","beta_par_squared"), function = _beta_par_squared,
                 units="g/cm**3")

    def _beta_perp_squared(field, data):
        return data["gas","density"]*data["gas","velocity_magnitude"]**2/clight/clight - data["gas","beta_par_squared"]
    ds.add_field(("gas","beta_perp_squared"), function = _beta_perp_squared,
                 units="g/cm**3")

    def _t_beta_par(field, data):
        return data["gas","kT"]*data["gas","beta_par"]
    ds.add_field(("gas","t_beta_par"), function = _t_beta_par,
                 units="keV*g/cm**3")

    def _t_sz(field, data):
        return data["gas","density"]*data["gas","kT"]
    ds.add_field(("gas","t_sz"), function = _t_sz,
                 units="keV*g/cm**3")

def generate_beta_par(L):
    def _beta_par(field, data):
        vpar = data["density"]*(data["velocity_x"]*L[0]+
                                data["velocity_y"]*L[1]+
                                data["velocity_z"]*L[2])
        return vpar/clight
    return _beta_par

class SZProjection(object):
    r""" Initialize a SZProjection object.

    Parameters
    ----------
    ds : ~yt.data_objects.static_output.Dataset 
        The dataset
    freqs : array_like
        The frequencies (in GHz) at which to compute the SZ spectral distortion.
    mue : float, optional
        Mean molecular weight for determining the electron number density.
    high_order : boolean, optional
        Should we calculate high-order moments of velocity and temperature?

    Examples
    --------
    >>> freqs = [90., 180., 240.]
    >>> szprj = SZProjection(ds, freqs, high_order=True)
    """
    def __init__(self, ds, freqs, mue=1.143, high_order=False):

        self.ds = ds
        self.num_freqs = len(freqs)
        self.high_order = high_order
        self.freqs = ds.arr(freqs, "GHz")
        self.mueinv = 1./mue
        self.xinit = hcgs*self.freqs.in_units("Hz")/(kboltz*Tcmb)
        self.freq_fields = ["%d_GHz" % (int(freq)) for freq in freqs]
        self.data = {}

        self.display_names = {}
        self.display_names["TeSZ"] = r"$\mathrm{T_e}$"
        self.display_names["Tau"] = r"$\mathrm{\tau}$"

        for f, field in zip(self.freqs, self.freq_fields):
            self.display_names[field] = r"$\mathrm{\Delta{I}_{%d\ GHz}}$" % int(f)

    def on_axis(self, axis, center="c", width=(1, "unitary"), nx=800, source=None):
        r""" Make an on-axis projection of the SZ signal.

        Parameters
        ----------
        axis : integer or string
            The axis of the simulation domain along which to make the SZprojection.
        center : A sequence of floats, a string, or a tuple.
            The coordinate of the center of the image. If set to 'c', 'center' or
            left blank, the plot is centered on the middle of the domain. If set to
            'max' or 'm', the center will be located at the maximum of the
            ('gas', 'density') field. Centering on the max or min of a specific
            field is supported by providing a tuple such as ("min","temperature") or
            ("max","dark_matter_density"). Units can be specified by passing in *center*
            as a tuple containing a coordinate and string unit name or by passing
            in a YTArray. If a list or unitless array is supplied, code units are
            assumed.
        width : tuple or a float.
            Width can have four different formats to support windows with variable
            x and y widths.  They are:

            ==================================     =======================
            format                                 example
            ==================================     =======================
            (float, string)                        (10,'kpc')
            ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
            float                                  0.2
            (float, float)                         (0.2, 0.3)
            ==================================     =======================

            For example, (10, 'kpc') requests a plot window that is 10 kiloparsecs
            wide in the x and y directions, ((10,'kpc'),(15,'kpc')) requests a
            window that is 10 kiloparsecs wide along the x axis and 15
            kiloparsecs wide along the y axis.  In the other two examples, code
            units are assumed, for example (0.2, 0.3) requests a plot that has an
            x width of 0.2 and a y width of 0.3 in code units.  If units are
            provided the resulting plot axis labels will use the supplied units.
        nx : integer, optional
            The dimensions on a side of the projection image.
        source : yt.data_objects.data_containers.YTSelectionContainer, optional
            If specified, this will be the data source used for selecting regions to project.

        Examples
        --------
        >>> szprj.on_axis("y", center="max", width=(1.0, "Mpc"), source=my_sphere)
        """
        axis = fix_axis(axis, self.ds)
        ctr, dctr = self.ds.coordinates.sanitize_center(center, axis)
        width = self.ds.coordinates.sanitize_width(axis, width, None)

        L = np.zeros(3)
        L[axis] = 1.0

        beta_par = generate_beta_par(L)
        self.ds.add_field(("gas","beta_par"), function=beta_par, units="g/cm**3")
        setup_sunyaev_zeldovich_fields(self.ds)
        proj = self.ds.proj("density", axis, center=ctr, data_source=source)
        frb = proj.to_frb(width[0], nx, height=width[1])
        dens = frb["density"]
        Te = frb["t_sz"]/dens
        bpar = frb["beta_par"]/dens
        omega1 = frb["t_squared"]/dens/(Te*Te) - 1.
        bperp2 = np.zeros((nx,nx))
        sigma1 = np.zeros((nx,nx))
        kappa1 = np.zeros((nx,nx))
        if self.high_order:
            bperp2 = frb["beta_perp_squared"]/dens
            sigma1 = frb["t_beta_par"]/dens/Te - bpar
            kappa1 = frb["beta_par_squared"]/dens - bpar*bpar
        tau = sigma_thompson*dens*self.mueinv/mh

        nx,ny = frb.buff_size
        self.bounds = frb.bounds
        self.dx = (frb.bounds[1]-frb.bounds[0])/nx
        self.dy = (frb.bounds[3]-frb.bounds[2])/ny
        self.nx = nx

        self._compute_intensity(np.array(tau), np.array(Te), np.array(bpar),
                                np.array(omega1), np.array(sigma1),
                                np.array(kappa1), np.array(bperp2))

        self.ds.field_info.pop(("gas","beta_par"))

    def off_axis(self, L, center="c", width=(1.0, "unitary"), depth=(1.0,"unitary"),
                 nx=800, nz=800, north_vector=None, no_ghost=False, source=None):
        r""" Make an off-axis projection of the SZ signal.

        Parameters
        ----------
        L : array_like
            The normal vector of the projection.
        center : A sequence of floats, a string, or a tuple.
            The coordinate of the center of the image. If set to 'c', 'center' or
            left blank, the plot is centered on the middle of the domain. If set to
            'max' or 'm', the center will be located at the maximum of the
            ('gas', 'density') field. Centering on the max or min of a specific
            field is supported by providing a tuple such as ("min","temperature") or
            ("max","dark_matter_density"). Units can be specified by passing in *center*
            as a tuple containing a coordinate and string unit name or by passing
            in a YTArray. If a list or unitless array is supplied, code units are
            assumed.
        width : tuple or a float.
            Width can have four different formats to support windows with variable
            x and y widths.  They are:

            ==================================     =======================
            format                                 example
            ==================================     =======================
            (float, string)                        (10,'kpc')
            ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
            float                                  0.2
            (float, float)                         (0.2, 0.3)
            ==================================     =======================

            For example, (10, 'kpc') requests a plot window that is 10 kiloparsecs
            wide in the x and y directions, ((10,'kpc'),(15,'kpc')) requests a
            window that is 10 kiloparsecs wide along the x axis and 15
            kiloparsecs wide along the y axis.  In the other two examples, code
            units are assumed, for example (0.2, 0.3) requests a plot that has an
            x width of 0.2 and a y width of 0.3 in code units.  If units are
            provided the resulting plot axis labels will use the supplied units.
        depth : A tuple or a float
            A tuple containing the depth to project through and the string
            key of the unit: (width, 'unit').  If set to a float, code units
            are assumed
        nx : integer, optional
            The dimensions on a side of the projection image.
        nz : integer, optional
            Deprecated, this is still in the function signature for API
            compatibility
        north_vector : a sequence of floats
            A vector defining the 'up' direction in the plot.  This
            option sets the orientation of the slicing plane.  If not
            set, an arbitrary grid-aligned north-vector is chosen.
        no_ghost: bool, optional
            Optimization option for off-axis cases. If True, homogenized bricks will
            extrapolate out from grid instead of interpolating from
            ghost zones that have to first be calculated.  This can
            lead to large speed improvements, but at a loss of
            accuracy/smoothness in resulting image.  The effects are
            less notable when the transfer function is smooth and
            broad. Default: True
        source : yt.data_objects.data_containers.YTSelectionContainer, optional
            If specified, this will be the data source used for selecting regions 
            to project.

        Examples
        --------
        >>> L = np.array([0.5, 1.0, 0.75])
        >>> szprj.off_axis(L, center="c", width=(2.0, "Mpc"))
        """
        wd = self.ds.coordinates.sanitize_width(L, width, depth)
        w = tuple(el.in_units('code_length').v for el in wd)
        ctr, dctr = self.ds.coordinates.sanitize_center(center, L)
        res = (nx, nx)

        if source is None:
            source = self.ds

        beta_par = generate_beta_par(L)
        self.ds.add_field(("gas","beta_par"), function=beta_par, units="g/cm**3")
        setup_sunyaev_zeldovich_fields(self.ds)

        dens = off_axis_projection(source, ctr, L, w, res, "density",
                                   north_vector=north_vector, no_ghost=no_ghost)
        Te = off_axis_projection(source, ctr, L, w, res, "t_sz",
                                 north_vector=north_vector, no_ghost=no_ghost)/dens
        bpar = off_axis_projection(source, ctr, L, w, res, "beta_par",
                                   north_vector=north_vector, no_ghost=no_ghost)/dens
        omega1 = off_axis_projection(source, ctr, L, w, res, "t_squared",
                                     north_vector=north_vector, no_ghost=no_ghost)/dens
        omega1 = omega1/(Te*Te) - 1.
        if self.high_order:
            bperp2 = off_axis_projection(source, ctr, L, w, res, "beta_perp_squared", 
                                         north_vector=north_vector, no_ghost=no_ghost)/dens
            sigma1 = off_axis_projection(source, ctr, L, w, res, "t_beta_par", 
                                         north_vector=north_vector, no_ghost=no_ghost)/dens
            sigma1 = sigma1/Te - bpar
            kappa1 = off_axis_projection(source, ctr, L, w, res, "beta_par_squared", 
                                         north_vector=north_vector, no_ghost=no_ghost)/dens
            kappa1 -= bpar
        else:
            bperp2 = np.zeros((nx,nx))
            sigma1 = np.zeros((nx,nx))
            kappa1 = np.zeros((nx,nx))
        tau = sigma_thompson*dens*self.mueinv/mh

        self.bounds = (-0.5*wd[0], 0.5*wd[0], -0.5*wd[1], 0.5*wd[1])
        self.dx = wd[0]/nx
        self.dy = wd[1]/nx
        self.nx = nx

        self._compute_intensity(np.array(tau), np.array(Te), np.array(bpar),
                                np.array(omega1), np.array(sigma1),
                                np.array(kappa1), np.array(bperp2))

        self.ds.field_info.pop(("gas","beta_par"))

    def _compute_intensity(self, tau, Te, bpar, omega1, sigma1, kappa1, bperp2):

        # Bad hack, but we get NaNs if we don't do something like this
        small_beta = np.abs(bpar) < 1.0e-20
        bpar[small_beta] = 1.0e-20

        comm = communication_system.communicators[-1]

        nx, ny = self.nx,self.nx
        signal = np.zeros((self.num_freqs,nx,ny))
        xo = np.zeros(self.num_freqs)

        k = int(0)

        start_i = comm.rank*nx//comm.size
        end_i = (comm.rank+1)*nx//comm.size

        pbar = get_pbar("Computing SZ signal.", nx*nx)

        for i in range(start_i, end_i):
            for j in range(ny):
                xo[:] = self.xinit[:]
                SZpack.compute_combo_means(xo, tau[i,j], Te[i,j],
                                           bpar[i,j], omega1[i,j],
                                           sigma1[i,j], kappa1[i,j], bperp2[i,j])
                signal[:,i,j] = xo[:]
                pbar.update(k)
                k += 1

        signal = comm.mpi_allreduce(signal)

        pbar.finish()

        for i, field in enumerate(self.freq_fields):
            self.data[field] = I0*self.xinit[i]**3*signal[i,:,:]
        self.data["Tau"] = self.ds.arr(tau, "dimensionless")
        self.data["TeSZ"] = self.ds.arr(Te, "keV")

    @parallel_root_only
    def write_fits(self, filename, sky_scale=None, sky_center=None, overwrite=True,
                   **kwargs):
        r""" Export images to a FITS file. Writes the SZ distortion in all
        specified frequencies as well as the mass-weighted temperature and the
        optical depth. Distance units are in kpc, unless *sky_center*
        and *scale* are specified. 

        Parameters
        ----------
        filename : string
            The name of the FITS file to be written. 
        sky_scale : tuple
            Conversion between an angle unit and a length unit, if sky
            coordinates are desired, e.g. (1.0, "arcsec/kpc")
        sky_center : tuple, optional
            The (RA, Dec) coordinate in degrees of the central pixel. Must
            be specified with *sky_scale*.
        overwrite : boolean, optional
            If the file already exists, do we overwrite?

        Additional keyword arguments are passed to
        :meth:`~astropy.io.fits.HDUList.writeto`.

        Examples
        --------
        >>> # This example just writes out a FITS file with kpc coords
        >>> szprj.write_fits("SZbullet.fits", overwrite=False)
        >>> # This example uses sky coords
        >>> sky_scale = (1., "arcsec/kpc") # One arcsec per kpc
        >>> sky_center = (30., 45., "deg")
        >>> szprj.write_fits("SZbullet.fits", sky_center=sky_center, sky_scale=sky_scale)
        """
        from yt.visualization.fits_image import FITSImageData

        dx = self.dx.in_units("kpc")
        dy = dx

        w = _astropy.pywcs.WCS(naxis=2)
        w.wcs.crpix = [0.5*(self.nx+1)]*2
        w.wcs.cdelt = [dx.v,dy.v]
        w.wcs.crval = [0.0,0.0]
        w.wcs.cunit = ["kpc"]*2
        w.wcs.ctype = ["LINEAR"]*2

        fib = FITSImageData(self.data, fields=self.data.keys(), wcs=w)
        if sky_scale is not None and sky_center is not None:
            fib.create_sky_wcs(sky_center, sky_scale)
        fib.writeto(filename, overwrite=overwrite, **kwargs)

    @parallel_root_only
    def write_png(self, filename_prefix, cmap_name=None,
                  axes_units="kpc", log_fields=None):
        r""" Export images to PNG files. Writes the SZ distortion in all
        specified frequencies as well as the mass-weighted temperature and the
        optical depth. Distance units are in kpc.

        Parameters
        ----------
        filename_prefix : string
            The prefix of the image filenames.

        Examples
        --------
        >>> szprj.write_png("SZsloshing")
        """
        if cmap_name is None:
            cmap_name = ytcfg.get("yt", "default_colormap")
        
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        if log_fields is None: log_fields = {}
        ticks_font = matplotlib.font_manager.FontProperties(family='serif',size=16)
        extent = tuple([bound.in_units(axes_units).value for bound in self.bounds])
        for field, image in self.items():
            data = image.copy()
            vmin, vmax = image.min(), image.max()
            negative = False
            crossover = False
            if vmin < 0 and vmax < 0:
                data *= -1
                negative = True
            if field in log_fields:
                log_field = log_fields[field]
            else:
                log_field = True
            if log_field:
                formatter = matplotlib.ticker.LogFormatterMathtext()
                norm = matplotlib.colors.LogNorm()
                if vmin < 0 and vmax > 0:
                    crossover = True
                    linthresh = min(vmax, -vmin)/100.
                    norm=matplotlib.colors.SymLogNorm(linthresh,
                                                      vmin=vmin, vmax=vmax)
            else:
                norm = None
                formatter = None
            filename = filename_prefix+"_"+field+".png"
            cbar_label = self.display_names[field]
            units = self.data[field].units.latex_representation()
            if units is not None and units != "":
                cbar_label += r'$\ \ ('+units+r')$'
            fig = plt.figure(figsize=(10.0,8.0))
            ax = fig.add_subplot(111)
            cax = ax.imshow(data.d, norm=norm, extent=extent, cmap=cmap_name, origin="lower")
            for label in ax.get_xticklabels():
                label.set_fontproperties(ticks_font)
            for label in ax.get_yticklabels():
                label.set_fontproperties(ticks_font)
            ax.set_xlabel(r"$\mathrm{x\ (%s)}$" % axes_units, fontsize=16)
            ax.set_ylabel(r"$\mathrm{y\ (%s)}$" % axes_units, fontsize=16)
            cbar = fig.colorbar(cax, format=formatter)
            cbar.ax.set_ylabel(cbar_label, fontsize=16)
            if negative:
                cbar.ax.set_yticklabels(["-"+label.get_text()
                                         for label in cbar.ax.get_yticklabels()])
            if crossover:
                yticks = list(-10**np.arange(np.floor(np.log10(-vmin)),
                                             np.rint(np.log10(linthresh))-1, -1)) + [0] + \
                         list(10**np.arange(np.rint(np.log10(linthresh)),
                                            np.ceil(np.log10(vmax))+1))
                cbar.set_ticks(yticks)
            for label in cbar.ax.get_yticklabels():
                label.set_fontproperties(ticks_font)
            fig.tight_layout()
            plt.savefig(filename)

    @parallel_root_only
    def write_hdf5(self, filename):
        r"""Export the set of S-Z fields to a set of HDF5 datasets.

        Parameters
        ----------
        filename : string
            This file will be opened in "write" mode.

        Examples
        --------
        >>> szprj.write_hdf5("SZsloshing.h5")
        """
        for field, data in self.items():
            data.write_hdf5(filename, dataset_name=field)

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def values(self):
        return self.data.values()

    def has_key(self, key):
        return key in self.data.keys()

    def __getitem__(self, key):
        return self.data[key]

    @property
    def shape(self):
        return (self.nx,self.nx)



"""
Tests for making line plots

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------
from nose.plugins.attrib import attr
from nose.tools import assert_raises

import yt
from yt.testing import assert_equal, fake_random_ds, ANSWER_TEST_TAG
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.line_plot import _validate_point

def setup():
    """Test specific setup."""
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"

def compare(ds, plot, test_prefix, test_name, decimals=12):
    def image_from_plot(filename_prefix):
        return plot.save(filename_prefix)

    image_from_plot.__name__ = "line_{}".format(test_prefix)
    test = GenericImageTest(ds, image_from_plot, decimals)
    test.prefix = test_prefix
    test.answer_name = test_name
    return test

@attr(ANSWER_TEST_TAG)
def test_line_plot():
    ds = fake_random_ds(4)
    fields = [field for field in ds.field_list if field[0] == 'stream']
    field_labels = {f:f[1] for f in fields}
    plot = yt.LinePlot(ds, fields, (0, 0, 0), (1, 1, 0), 1000,
                       field_labels=field_labels)
    plot.annotate_legend(fields[0])
    plot.annotate_legend(fields[1])
    plot.set_x_unit('cm')
    plot.set_unit(fields[0], 'kg/cm**3')
    plot.annotate_title(fields[0], "Density Plot")
    yield compare(ds, plot, test_prefix="answers_line_plot",
                  test_name="answers_line_plot")

@attr(ANSWER_TEST_TAG)
def test_multi_line_plot():
    ds = fake_random_ds(4)
    fields = [field for field in ds.field_list if field[0] == 'stream']
    field_labels = {f: f[1] for f in fields}
    lines = []
    lines.append(yt.LineBuffer(ds, [0.25, 0, 0], [0.25, 1, 0], 100,
                               label='x = 0.5'))
    lines.append(yt.LineBuffer(ds, [0.5, 0, 0], [0.5, 1, 0], 100,
                               label='x = 0.5'))
    plot = yt.LinePlot.from_lines(ds, fields, lines, field_labels=field_labels)
    plot.annotate_legend(fields[0])
    plot.annotate_legend(fields[1])
    yield compare(ds, plot, test_prefix="answers_multi_line_plot",
                  test_name="answers_multi_line_plot")

def test_line_buffer():
    ds = fake_random_ds(32)
    lb = yt.LineBuffer(ds, (0, 0, 0), (1, 1, 1), 512, label='diag')
    lb['density']
    lb['velocity_x']
    assert_equal(lb['density'].size, 512)
    lb['density'] = 0
    assert_equal(lb['density'], 0)
    assert_equal(set(lb.keys()), set(['density', 'velocity_x']))
    del lb['velocity_x']
    assert_equal(set(lb.keys()), set(['density']))

def test_validate_point():
    ds = fake_random_ds(3)
    with assert_raises(RuntimeError) as ex:
        _validate_point(0, ds, start=True)
    assert_equal(str(ex.exception), 'Input point must be array-like')

    with assert_raises(RuntimeError) as ex:
        _validate_point(ds.arr([[0], [1]], 'code_length'), ds, start=True)
    assert_equal(str(ex.exception), 'Input point must be a 1D array')

    with assert_raises(RuntimeError) as ex:
        _validate_point(ds.arr([0, 1], 'code_length'), ds, start=True)
    assert_equal(str(ex.exception),
                 'Input point must have an element for each dimension')

    ds = fake_random_ds([32, 32, 1])
    _validate_point(ds.arr([0, 1], 'code_length'), ds, start=True)
    _validate_point(ds.arr([0, 1], 'code_length'), ds)
